# Copyright(c) 2000-2004 Association of Universities for Research in Astronomy, Inc.

procedure gemqa(inimages,finalint,finaliq,piconfig,pitarget)

# Add data quality assessemnt keywords to reduced images
# 
# Version  Oct 12, 2001 IJ, v1.2 release
#          Sept 20,2002 IJ, v1.4 release

char    inimages  {prompt="Images to be updated"}  # OLDP-1-input-primary-update
char    finalint  {"",min="pass|fail|unknown",prompt="Final integration time (pass|fail)"}  # OLDP-1
char    finaliq   {"",min="pass|fail|unknown",prompt="Final image quality (pass|fail)"}  # OLDP-1
char    piconfig  {"",min="pass|fail|unknown",prompt="Used correct instrument configuration (pass|fail)"}  # OLDP-1
char    pitarget  {"",min="pass|fail|unknown",prompt="Observed correct target (pass|fail)"}  # OLDP-1
char    sci_ext   {"0",prompt="If MEF, name or number of extension to update"}  # OLDP-3
char    logfile   {"gemqa.log",prompt="Name of log file"}  # OLDP-1
bool    verbose   {no,prompt="Verbose actions"}  # OLDP-4
int     status    {0,prompt="Exit status (0=good)"}  # OLDP-4
struct* scanfile  {"",prompt="Internal use only"}  # OLDP-4

begin

    # declare local variables
    char    l_inimages, l_finalint, l_finaliq, l_piconfig, l_pitarget
    char    l_sci_ext, l_logfile
    char    tmpfile1, tmpfile2, img
    char    projID,newprojID,prefix,suffix,datalabel,obsID
    bool    l_verbose
    struct  l_struct

    status=0
    scanfile=""
    cache ("imgets")

    # Set local variables to input parameters
    l_inimages=inimages 
    l_finalint=finalint ; l_finaliq=finaliq ; l_piconfig=piconfig
    l_pitarget=pitarget
    l_sci_ext=sci_ext ; l_logfile=logfile
    l_verbose=verbose

    # Check for package log file or user-defined log file
    if ((l_logfile=="") || (l_logfile==" ")) {
	l_logfile="gemqa.log"
	printlog("WARNING - GEMQA: No logfile defined. Using default file \
	    gemqa.log",logfile=l_logfile, verbose+)
    }

    # Open log file
    date | scan(l_struct)
    printlog("--------------------------------------------------------------\
        --------------",logfile=l_logfile, verbose=l_verbose)
    printlog("GEMQA -- "//l_struct, logfile=l_logfile, verbose=l_verbose)
    printlog(" ",logfile=l_logfile, verbose=l_verbose)

    # Make temporary files
    tmpfile1 = mktemp("tmpfl1")
    tmpfile2 = mktemp("tmpfl2")

    #----------------------------------------------------------------------
    # Put all images in a temporary image list file: tmpfile
    if (substr(l_inimages,1,1)=="@") 
	type(substr(l_inimages,2,strlen(l_inimages)), > tmpfile1)
    else 
	files(l_inimages,sort-, > tmpfile1)


    # Verify that input images actually exist. (Redundant for * though)
    # and remove .fits if present
    scanfile = tmpfile1
    while (fscan(scanfile,img) != EOF) {

	if (imaccess(img)) {
	    if (substr(img,strlen(img)-4,strlen(img)) == ".fits") 
		img=substr(img,1,(strlen(img)-5))

	    imgets(img//"[0]","EXTEND", >& "dev$null")
	    if (imgets.value =="T" ) {
		if (imaccess(img//"["//l_sci_ext//"]"))
		    print(img//"["//l_sci_ext//"]", >> tmpfile2)
		else {
		    printlog("ERROR - GEMQA: image "//img//l_sci_ext//\
		        " does not exist",logfile=l_logfile, verbose+)
		    status=1
		    goto clean
		}
	    } else
		print(img, >> tmpfile2)

	} else {
	    printlog("ERROR - GEMQA: image "//img//" does not exist.",
		logfile=l_logfile, verbose+)
	    status=1
	    goto clean
	}
    }
    scanfile=""

    printlog("Updating images:",logfile=l_logfile, verbose=l_verbose)
    if (l_verbose) type(tmpfile2)
    type(tmpfile2, >> l_logfile)

    printlog("FINALINT = "//l_finalint,logfile=l_logfile, verbose=l_verbose)
    printlog("FINALIQ  = "//l_finaliq,logfile=l_logfile, verbose=l_verbose)
    printlog("PICONFIG = "//l_piconfig,logfile=l_logfile, verbose=l_verbose)
    printlog("PITARGET = "//l_pitarget,logfile=l_logfile, verbose=l_verbose)

    #-------------------------------------------------------------------------
    # Update the headers
    scanfile=tmpfile2
    while (fscan(scanfile,img)!=EOF) {

	# Make DATALAB a unique identifier
	
	# Parse the image name. get prefix and suffix	
	hselect(img, "DATALAB", yes) | scan(datalabel)
	prefix=""; suffix=""
	if ( strstr("N",datalabel) != 0 ) {
	    if ( stridx("N",img) > 1 )
	        prefix = "-"//substr(img,1,stridx("N",img)-1)
	} else if ( strstr("S",datalabel) != 0 ) {
	    if ( stridx("S",img) > 1 )
	        prefix = "-"//substr(img,1,stridx("S",img)-1)
	} else {
	    printlog("ERROR - GEMQA: project ID, "//datalabel//", is invalid",
	        logfile=l_logfile, verbose+)
	    status=1
	    goto clean
	}
	if ( stridx("_",img) != 0 ) {
	    suffix = "-"//substr(img, stridx("_",img)+1, strlen(img)-3 )
	    print(suffix) | translit("STDIN","_","-",del-,col-) | scan(suffix)
	}
	
	# Parse the datalabel to extract the 'obsID-XXX' string; rebuild string
	hselect(img, "OBSID", yes) | scan(obsID)
	projID = substr(datalabel, 1, strlen(obsID)+4)
	datalabel = projID//prefix//suffix
	hedit(img,"DATALAB",datalabel,add+,addonly-,delete-,verify-,
	    show-,update+)	
	
	# Edit the other keywords
	gemhedit(img,"FINALINT",l_finalint,"Requested total integration time")
	gemhedit(img,"FINALIQ",l_finaliq,"Requested final image quality")
	gemhedit(img,"PICONFIG",l_piconfig,"Requested instrument configuration")
	gemhedit(img,"PITARGET",l_pitarget,"Requested target")

    }
    scanfile=""

    #------------------------------------------------------------------------

    # Clean up
clean:
    if (status==0)
	printlog("GEMQA exit status:  good",logfile=l_logfile,verbose=l_verbose)
    printlog("--------------------------------------------------------------\
        --------------",logfile=l_logfile, verbose=l_verbose)
    delete(tmpfile1//","//tmpfile2,ver-, >& "dev$null")
end

