# Copyright(c) 2002-2004 Association of Universities for Research in Astronomy, Inc.

procedure gprepare(inimages)

# Takes GMOS raw data (all modes), 3 SCI extensions and creates
# VAR and DQs for each if desired.  Updated header keywords as necessary
#
# Variance frame is generated as:
#        var = (readnoise/gain)^2 + max(data,0.)/gain
# DQ frame is generated as using BPM to set 1 bit and saturation 
#  level to set 4 bit
# MDF may be attached 
#
# Version   Feb 28, 2002  CW,BM,IJ  v1.3 release
#           Jul 7, 2002   IJ  N&S support, bugfix comma-sep lists
#           Aug 9, 2002   IJ add pixscale to headers to support imcoadd
#           Aug 12, 2002  IJ fixed logfile/l_logfile use
#           Aug 19, 2002  IJ more robust for incomplete headers/incorrect input
#                            warn if processing spec data w/ fl_addmdf-
#                            mirror-> MIRROR
#           Aug 26, 2002  IJ parameter encoding
#           Sept 20, 2002 IJ v1.4 release
#           Oct 14, 2002  IJ don't use ijk from within script
#           Mar 20, 2003  BM add different pixscale's for GMOS-N,S
#           May 9, 2003   IJ changed INSTRUME logic to support old GMOS-N data
#           Aug 26, 2003  KL IRAF2.12 - new parameter, addonly, in hedit

char inimages      {"",prompt="Input GMOS images or list"}  # OLDP-1-input-primary-single-prefix=g
char rawpath       {"",prompt="Path for input raw images"}  # OLDP-4
char outimages     {"",prompt="Output images or list"}      # OLDP-1-output
char outpref       {"g",prompt="Prefix for output images"}  # OLDP-4
bool fl_vardq      {no,prompt="Create variance and data quality frames"} # OLDP-3
bool fl_addmdf     {no,prompt="Add Mask Definition File (LONGSLIT/MOS/IFU modes)"} # OLDP-2
char sci_ext       {"SCI",prompt="Name of science extension"}                  # OLDP-3
char var_ext       {"VAR",prompt="Name of variance extension"}                 # OLDP-3
char dq_ext        {"DQ",prompt="Name of data quality extension"}              # OLDP-3
char key_mdf       {"MASKNAME",prompt="Header keyword for the Mask Definition File"} # OLDP-3
char mdffile       {"",prompt="MDF file to use if keyword not found"}          # OLDP-2-input
char mdfdir        {"gmos$data/", prompt="MDF database directory"}             # OLDP-2
char bpm           {"",prompt="Bad Pixel Mask filename"}                       # OLDP-3-input
char gaindb        {"gmos$data/gmosamps.dat",prompt="Database with gain data"} # OLDP-3
int  sat           {65000,prompt="Saturation level in raw images (ADU)"}       # OLDP-3
char key_ron       {"RDNOISE",prompt="Header keyword for readout noise"}       # OLDP-3
char key_gain      {"GAIN",prompt="Header keyword for gain (e-/ADU)"}          # OLDP-3
real ron           {3.5,prompt="Readout noise in electrons"}                   # OLDP-3
real gain          {2.2,prompt="Gain in e-/ADU"}                               # OLDP-3
char key_dsec      {"DATASEC",prompt="Header keyword for data section"}        # OLDP-3
char logfile       {"",prompt="Logfile"}                                       # OLDP-1
bool verbose       {yes,prompt="Verbose?"}                                     # OLDP-4
int  status        {0,prompt="Exit status (0=good)"}                           # OLDP-4
struct* scanfile   {"",prompt="Internal use only"}                             # OLDP-4

begin

    # Define local variables:
    char   l_inimages, l_outimages, l_outpref, l_bpm, l_logfile, l_sci_ext
    char   l_var_ext, l_dq_ext, l_key_ron, l_key_gain, l_key_dsec, l_key_mdf
    char   l_mdffile,l_mdfdir, l_rawpath, l_temp, l_gaindb
    bool   l_fl_vardq, l_verbose, l_fl_addmdf
    real   l_ron, l_gain, pixscale[2]
    int    l_sat, maxfiles, maxscext, nbad
    struct l_struct
    char   img, filelist, inimg[200], outimg[200], varexp, amatch
    char   dqexp, test, fxvexp, fxdexp, tmpv[50], tmpd[50], tmpsci, tmpdq
    char   sci_tsec, bpm_tsec, suf, sci_dsec, bpm_dsec, tsec, mdfname[200]
    char   obsmode[200], obs_bpm
    bool   addmdf[200]
    int    ninp, len, nout, i, next_bpm, next_inp, next_out, n_j, next_sci, n_i
    int    n_var[50], n_dq[50], varpos, dqpos, temp, l_test, ybin, dum, inst
    int    n_test,n_itest
    char   l_dir

    # Make temporary files
    filelist = mktemp("tmpfile")
 
    # Set local variable values:
    l_inimages=inimages; l_outimages=outimages; l_outpref=outpref; l_bpm=bpm
    l_logfile=logfile; l_sci_ext=sci_ext; l_var_ext=var_ext; l_dq_ext=dq_ext
    l_key_ron=key_ron; l_key_gain=key_gain; l_key_dsec=key_dsec
    l_key_mdf=key_mdf
    l_mdffile=mdffile; l_mdfdir = mdfdir ; l_rawpath=rawpath
    l_fl_vardq=fl_vardq; l_verbose=verbose; l_fl_addmdf=fl_addmdf
    l_ron=ron; l_gain=gain ; l_sat=sat ; l_gaindb=gaindb

    l_dir=""

    status=0
    maxfiles=200
    maxscext=50
    next_bpm=0
    obs_bpm=""
    pixscale[1]=0.0727
    pixscale[2]=0.073

    # Keep task parameters from changing from the outside
    cache("imgets","gimverify","fparse")

    # Test the logfile:
    if (l_logfile == "" || stridx(" ",l_logfile)>0) {
	l_logfile = gmos.logfile
	if (l_logfile == "" || stridx(" ",l_logfile)>0) {
            l_logfile = "gmos.log"
            printlog("WARNING - GPREPARE: Both gprepare.logfile and gmos.logfile fields are empty",
	        l_logfile,yes)
            printlog("                    Using default file gmos.log",
	        l_logfile,yes)
	}
    }

    # Check the rawpath name for a final /
    if (substr(l_rawpath,(strlen(l_rawpath)),(strlen(l_rawpath))) != "/")
        l_rawpath=l_rawpath//"/"
    if (l_rawpath=="/" || l_rawpath==" ")
        l_rawpath=""

    # Start logging
    date | scan(l_struct)
    printlog("----------------------------------------------------------------------------",
        l_logfile,l_verbose)
    printlog("GPREPARE -- "//l_struct,l_logfile,l_verbose)
    printlog("",l_logfile,l_verbose)
    printlog("Input list    = "//l_inimages,l_logfile,l_verbose)
    printlog("Output list   = "//l_outimages,l_logfile,l_verbose)
    printlog("Output prefix = "//l_outpref,l_logfile,l_verbose)
    printlog("Raw path      = "//l_rawpath,l_logfile,l_verbose)
    printlog("VAR/DQ planes = "//l_fl_vardq,l_logfile,l_verbose)
    printlog("Add MDF       = "//l_fl_addmdf,l_logfile,l_verbose)
    if (l_fl_addmdf)
        printlog("Input MDF in case header keyword not found = "//l_mdffile,
	    l_logfile,l_verbose)
    printlog("",l_logfile,l_verbose)

    # Test the SCI extension name:
    if (l_sci_ext == "" || stridx(" ",l_sci_ext)>0) {
	printlog("ERROR - GPREPARE: Science extension name sci_ext is not defined",
	    l_logfile,yes)
	goto crash
    }               

    # Test the VAR and DQ extension names, the RDNOISE, GAIN and DATASEC keywords
    if (l_fl_vardq) {
	if (l_var_ext == "" || stridx(" ",l_var_ext) >0 || l_dq_ext == "" || stridx(" ",l_dq_ext)>0 ) {
            printlog ("ERROR - GPREPARE: Variance and/or data quality extension names var_ext/dq_ext not defined",
	        l_logfile,yes)
            goto crash
	}

	if (l_key_ron == "" || stridx(" ",l_key_ron)>0) {
            printlog ("ERROR - GPREPARE: Readout noise keyword key_ron is not defined",
                l_logfile,yes)
            goto crash
	}

	if (l_key_gain == "" || stridx(" ",l_key_gain)>0) {
            printlog ("ERROR - GPREPARE: Gain keyword key_gain is not defined",
                l_logfile,yes)
            goto crash
	}

	if (l_key_dsec == "" || stridx(" ",l_key_dsec)>0) {
            printlog ("ERROR - GPREPARE: Data section keyword key_dsec is not defined",
                l_logfile,yes)
            goto crash
	}
    }    

    # Test the MDF related parameters
    if (l_fl_addmdf) {
	if ((l_key_mdf == "" || stridx(" ",l_key_mdf)>0 ) && (l_mdffile == "" || stridx(" ",l_mdffile)>0)) {
	    printlog ("ERROR - GPREPARE: Neither the MDF keyword key_mdf or the",
	        l_logfile,yes)
	    printlog ("                  MDF filename mdffile are defined",
	        l_logfile,yes)
	    goto crash
	}
    }        

    # Test gaindb
    if (l_gaindb != "" && stridx(" ",l_gaindb)==0) {
	if (!access(l_gaindb)) {
            printlog("ERROR - GPREPARE: Gain database not found",
	        l_logfile,yes)
            goto crash
	}
    }

    #-------------------------------------------------------------------
    # Load up input name list: @list, * and ?, comma separated

    if (l_inimages == "" || stridx(" ",l_inimages)>0) {
	printlog("ERROR - GPREPARE: Input file not specified",l_logfile,yes)
	goto crash
    }

    # Test for @filelist
    if (substr(l_inimages,1,1) == "@" && !access(substr(l_inimages,2,strlen(l_inimages))) ) {
        printlog("ERROR - GPREPARE: Input list "//substr(l_inimages,2,strlen(l_inimages))//" does not exist",
	    l_logfile,yes)
        goto crash
    }

    # parse wildcard and comma-separated lists
    if (substr(l_inimages,1,1)=="@")  {
        scanfile=substr(l_inimages,2,strlen(l_inimages))
        while(fscan(scanfile,l_temp) != EOF) {
	    files(l_rawpath//l_temp, >> filelist)
        }
    } else {
        if (stridx(",",l_inimages)==0) 
	    files(l_rawpath//l_inimages, > filelist)
        else {
	    l_test=9999
	    while (l_test!=0) {
	        l_test=stridx(",",l_inimages)
	        if (l_test>0)
	            files(l_rawpath//substr(l_inimages,1,l_test-1), >> filelist)
	        else
	            files(l_rawpath//l_inimages, >> filelist)
	        l_inimages=substr(l_inimages,l_test+1,strlen(l_inimages))
	    }
        }
    }  
    scanfile=""
    scanfile=filelist

    #--------------------------------------------------------------------
    # check the input images

    ninp = 0
    nbad=0
    while (fscan(scanfile, img) != EOF) {
        # split off directory path
        fparse(img,verbose-)
        img=fparse.root
        l_dir=fparse.directory

        # Must be there, but cannot be GEIS or OIF
	gimverify(l_dir//img)
	if (gimverify.status==0) {
	    ninp = ninp+1
	    if (ninp > maxfiles) {
                printlog("ERROR - GPREPARE: Maximum number of input images exceeded",
                    l_logfile,verbose+)
                goto crash
	    }
	    inimg[ninp] = gimverify.outname//".fits"  # now has path if relevant
	    imgets(l_dir//img//"[0]","GPREPARE", >>& "dev$null")
	    if (imgets.value != "0") {
                printlog("ERROR - GPREPARE:  Image "//l_dir//img//" already GPREPAPREd",
                    l_logfile,verbose+)
                nbad+=1
	    }
	} else if (gimverify.status==1) {
	    printlog("ERROR - GPREPARE: Input image "//l_dir//img//" does not exist",
                l_logfile,yes)
	    nbad+=1
	} else {
	    printlog("ERROR - GPREPARE: Input image "//l_dir//img//" is not MEF",
                l_logfile,yes)
	    nbad+=1
	}
    } # end while

    if (nbad!=0) {
        printlog("ERROR - GPREPARE: "//nbad//" input files do not exist, are the wrong type, or",
	    l_logfile,verbose+)
        printlog("                  are already GPREPAREd",
	    l_logfile,verbose+)
        goto crash
    }
    if (ninp==0) {
        printlog("ERROR - GPREPARE: No input images meet wildcard criteria",
	    l_logfile,verbose+)
        goto crash
    }

    # inimg[ninp] now contains the input images incl. the directory path
    for (n_i=1;n_i<=ninp;n_i+=1) {
        fparse (inimg[n_i],verbose-)
        img=fparse.root
        l_dir=fparse.directory

        # Identify the OBSMODE and set the keyword value. Note that if it cannot 
        # find the keyword will automatically set as IMAGE
	imgets (l_dir//img//"[0]", "MASKTYP", >>& "dev$null")    
	if (imgets.value == "0")
	    obsmode[n_i] = "IMAGE"
	else if (imgets.value == "-1")
	    obsmode[n_i] = "IFU"
	else if (imgets.value == "1" )
	    obsmode[n_i] = "MOS"
	else {
	    # If cannot determine what it is, assume it is IMAGE instead of 
	    # crashing
	    printlog ("WARNING - GPREPARE: The MASKTYP keyword is not recognizable - assuming  OBSMODE=IMAGE",
	        l_logfile,verbose+)
	    obsmode[n_i] = "IMAGE"
	}

        # If OBSMODE = MOS, check the MASKNAME for longslit
	amatch=""
	if (obsmode[n_i] == "MOS") {
	    imgets (l_dir//img//"[0]", "MASKNAME", >>& "dev$null")
	    print(imgets.value) | match("arcsec","STDIN",stop-) | \
	        match("NS","STDIN",stop+) | scan(amatch)
	    if (strlen(amatch)>0)
                obsmode[n_i] = "LONGSLIT"
	}

        # If adding the MDF, test if the image is of a valid format to receive 
	# an MDF (IMAGEs do not have MDFs)
	if (l_fl_addmdf) {
	    if (obsmode[n_i] == "IMAGE")
                addmdf[n_i] = no
	    else
                addmdf[n_i] = yes
	}       
        # Warn if not adding MDF to a spectroscopic observation
	if (!l_fl_addmdf && obsmode[n_i]!="IMAGE")
	    printlog("WARNING - GPREPARE: Spectroscopic frame "//img//" will have no MDF, use fl_addmdf=yes",
	        l_logfile,l_verbose)

        # Double check the grating choice
        # Checking for it here will allow the MDF to be attached with 
	# OBSMODE=IMAGE
	imgets (l_dir//img//"[0]", "GRATING", >>& "dev$null")
	if (imgets.value=="MIRROR" && obsmode[n_i]!="IMAGE") {
	    printlog("WARNING - GPREPARE: Mask or IFU used without grating, setting OBSMODE=IMAGE",
	        l_logfile,verbose+)
	    obsmode[n_i] = "IMAGE"
	}    

    } # end of for loop over input images -- while(img) loop

    scanfile=""
    delete(filelist,ver-, >>& "dev$null")

    #--------------------------------------------------------------------
    # Output name list: can be empty if prefix is defined, @list or 
    # comma-separated list

    if (stridx(" ",l_outimages)>0) l_outimages = ""

    if (substr(l_outimages,1,1) == "@" && !access(substr(l_outimages,2,strlen(l_outimages))) ) {
        printlog("ERROR - GPREPARE: File "//substr(l_outimages,2,strlen(l_outimages))//" does not exist",
	    l_logfile,yes)
        goto crash
    }

    files(l_outimages,sort-, > filelist)
    scanfile=filelist
    nout = 0

    # If empty string, prefix must be defined
    if (l_outimages == "") {
	if (l_outpref == "" || stridx(" ",l_outpref)>0 ) {
	  printlog("ERROR - GPREPARE: Neither output name nor output prefix is defined",
	      l_logfile,verbose+)
	  goto crash
	}
	nout = ninp
	n_i=1

	while (n_i<=nout) {
            fparse(inimg[n_i],verbose=no)
            outimg[n_i] = l_outpref//fparse.root//".fits"        
            gimverify(outimg[n_i])
            if (gimverify.status!=1) {
                printlog("ERROR - GPREPARE: Output image "//outimg[n_i]//" already exists",
        	    l_logfile, verbose+)
                nbad+=1
            }
            n_i=n_i+1
	} # end of while(n_i) loop            

    } else {
	while (fscan(scanfile, img) != EOF) {
            gimverify(img)
            if (gimverify.status!=1) {
        	printlog("ERROR - GPREPARE: Output image "//img//" already exists",
        	    l_logfile,verbose+)
        	nbad+=1
            }         
            nout = nout+1
            if (nout > maxfiles) {
        	printlog("ERROR - GPREPARE: Maximum number of output images exceeded",
        	    l_logfile,verbose+)
        	goto crash
            }
            outimg[nout] = gimverify.outname//".fits"
	} # end of while (img) loop

    } # end else

    scanfile=""
    delete(filelist,ver-, >>& "dev$null")

    if (nbad>0)
        goto crash

    # Input and output number must be the same, if output names are defined
    if (ninp!=nout && l_outimages!="") {
	printlog("ERROR - GPREPARE: Number of input and output images are not the same",
            l_logfile,yes)
	goto crash
    }

    # Now check for existence of Bad Pixel Mask 
    if (l_fl_vardq && !imaccess(l_bpm) && l_bpm!="" && stridx(" ",l_bpm)<=0) {
	printlog("WARNING - GPREPARE: Bad Pixel Mask "//l_bpm//" not found",
	    l_logfile,yes)
	printlog("                    Only saturated pixels will be flagged",
	    l_logfile,yes)
	l_bpm=""
	next_bpm=0
    } else if (l_fl_vardq && (l_bpm == "" || stridx(" ",l_bpm)>0)) {
	printlog("WARNING - GPREPARE: Bad Pixel Mask filename is an empty string",
             l_logfile,yes)
	printlog("                    Only saturated pixels will be flagged",
             l_logfile,yes)
	l_bpm=""
	next_bpm=0
    }

    # Check if BPM is a MEF and keep the number of extensions for later use
    if (l_fl_vardq && l_bpm != "") {
        gimverify (l_bpm)
        if (gimverify.status == 0) {
	    # Must be .fits (gimverify strips the .fits if it was there, so we have to put it back)
	    l_bpm = gimverify.outname//".fits"
        } else if (gimverify.status == 1) {
	    printlog("ERROR - GPREPARE: BPM "//l_bpm//" does not exist",
	        l_logfile,yes)
	    goto crash
        } else {
	    printlog("ERROR - GPREPARE: BPM "//l_bpm//" is not a MEF file.",
	        l_logfile,yes)
	    goto crash
        }

        # Now get the number of science extensions. We can change this back to 
	# imget(NEXTEND) as we are keeping the keyword in the PHU (it is 
	# updated at the end of the task).
        # Assumes all extensions are to be used as BPM, IJ

	imgets(l_bpm//"[0]","NEXTEND", >>& "dev$null")
	next_bpm = int(imgets.value)

        # Get the OBSMODE of the BPM
	imgets (l_bpm//"[0]", "OBSMODE", >>& "dev$null")
	obs_bpm = imgets.value
    }

    # Define the expressions for imexpr
    varexp ="((a/b)**2. + max(c,0.0)/b)"
    dqexp = "((a > b) ? 4 : 0)"

    #---------------------------------------------------------------
    # Start making the output images
    n_i=1

    while (n_i<=ninp) {
	img = inimg[n_i]

	# Which instrument?
	imgets(img//"[0]","INSTRUME", >>& "dev$null")
	if (imgets.value == "0") {
	    printlog ("ERROR - GPREPARE: Instrument keyword not found.",
	        l_logfile, verbose+)
	    goto crash
	}
	inst=1  # Default is GMOS-N, support for old data
	if (imgets.value == "GMOS-S")
            inst=2
	printlog("Input  "//img//"   Output  "//outimg[n_i],l_logfile,l_verbose)

        # Raw images do not have NEXTEND in the header, so get the number of 
	# extensions from fxheader
	# Most peculiar bug?, need the extra fxhead to make sure the lines get 
	# counted correctly???, IJ
	fxhead (img,format_file="",long_head-,count_lines-, >>& "dev$null")
	fxhead (img,format_file="",long_head-,count_lines-) | \
	    count ("STDIN") | scan (next_inp)
	fxhead (img,format_file="",long_head-,count_lines-) | \
	    fields ("STDIN","1",lines=str(next_inp),quit_if_miss-,
	    print_file-) | scan (next_inp)
        # Test if the MDF keyword exists in the header, check if it is the 
	# same as the input file. 
	if (l_fl_addmdf && addmdf[n_i]) {
            imgets (img//"[0]",l_key_mdf, >>& "dev$null")
            if (imgets.value == "0") {
        	printlog ("GPREPARE: Using MDF defined in the parameter list "//l_mdffile,
		    l_logfile,verbose+)
        	mdfname[n_i] = l_mdffile
            } else {
        	mdfname[n_i] = imgets.value
        	printlog ("GPREPARE: Using MDF defined in the header "//mdfname[n_i],
		    l_logfile, verbose+)
            } 
            len = strlen(mdfname[n_i])
            suf = substr(mdfname[n_i],len-4,len)
            if (suf !=".fits" )
                mdfname[n_i] = mdfname[n_i]//".fits"
            if (access(mdfname[n_i]))
                printlog("GPREPARE: Taking MDF from the current directory.",
        	    l_logfile,l_verbose)
            else {
                printlog("GPREPARE: Taking MDF from directory "//mdfdir,
        	    l_logfile,l_verbose)
                mdfname[n_i]=l_mdfdir//mdfname[n_i]
            }
            if (mdfname[n_i] == "" || stridx(" ",mdfname[n_i])>0) {
        	printlog ("ERROR - GPREPARE: The MDF filename is not defined.",
		    l_logfile, verbose+)
        	goto crash
            }
            if (!access(mdfname[n_i])) {
        	printlog ("ERROR - GPREPARE: The MDF file does not exist.",
		    l_logfile, verbose+)
        	goto crash
            }

	}   # end of fl_addmdf 

        #-------------------------
        #If no problems found, creates the new image and insert the MDF 
	#if requested:
	copy (img, outimg[n_i], ver-, >>& "dev$null")

        # Update EXTNAME and EXTVER, assuming all extensions but the MDF 
	# are SCI
	n_j=1
	next_sci = 0
	while (n_j<=next_inp) {
            keypar (img//"["//n_j//"]", "EXTNAME", silent-, >>& "dev$null")
            if (keypar.value == "MDF")
        	addmdf[n_i] = no
            else {
        	hedit(outimg[n_i]//"["//n_j//"]","EXTNAME",l_sci_ext,add+,
		    addonly-,delete-,verify-,show-,update+)
        	hedit(outimg[n_i]//"["//n_j//"]","EXTVER", n_j,add+,addonly-,
	            delete-,verify-,show-,update+)        
        	next_sci = next_sci + 1
            } 
            n_j = n_j+1
	}
	flpr # necessary to work around an IRAF bug 

	if (next_sci > maxscext) {
	    printlog("ERROR - GPREPARE: Maximum number of "//l_sci_ext//" extensions exceeded.",
	        l_logfile,verbose+)
	    imdelete (outimg[n_i],ver-, >>& "dev$null")
	    goto crash
	}

	if (l_fl_addmdf && addmdf[n_i]) {
            fxinsert (mdfname[n_i],outimg[n_i]//"["//next_inp//"]",
	        groups="1",verbose-, >>& "dev$null")
            next_inp = next_inp+1
            # Clean old MDFs with double EXTNAME
            n_test=1
            tprint(outimg[n_i]//"["//next_inp//"]",prpar+,prdat-,option="plain") | \
            match("EXTNAME","STDIN",stop-) | count("STDIN") | scan(n_test)
            for(n_itest=1;n_itest<n_test;n_itest+=1)
                 thedit(outimg[n_i]//"["//next_inp//"]","EXTNAME","",delete+,show-)
            parkey ("MDF", outimg[n_i]//"["//next_inp//"]", "EXTNAME", add+)
            parkey ("1", outimg[n_i]//"["//next_inp//"]","EXTVER", add+)
	}    
	next_out = next_inp

        # Update gain and readnoise
	gemhedit(outimg[n_i]//"[0]","NSCIEXT",next_sci,
	    "Number of science extensions")
	if (access(l_gaindb)) {
            ggain(outimg[n_i],gaindb=l_gaindb,logfile=l_logfile,
	        key_gain=l_key_gain,key_ron=l_key_ron,gain=l_gain,ron=l_ron,
		fl_update+,fl_mult-,verbose=no,sci_ext=l_sci_ext,
		var_ext=l_var_ext)
            if (ggain.status != 0) {
        	printlog("ERROR - GPREPARE: error in GGAIN",l_logfile,yes)
        	goto crash
            }
	}

        #--------------------------------
        # Creates VAR and DQ if requested    
	if (l_fl_vardq) {

            if (l_verbose) print("Computing VAR and DQ planes for image "//outimg[n_i])

            # SCI should be the same number in the BPM
            if (next_sci != next_bpm && l_bpm != "") {
        	printlog("ERROR - GPREPARE: Number of "//l_sci_ext//" extensions in "//img,
		    l_logfile,verbose+)             
        	printlog("                  is not the same as in the Bad Pixel Mask." ,
		    l_logfile,verbose+)
        	imdelete (outimg[n_i],ver-, >>& "dev$null")
        	goto crash
            }

            # Using an IMAGE mode BPM to flag IFU/MOS/LONGSLIT is not quite right
            if (l_bpm != "") {
        	if (obs_bpm == "IMAGE" && obsmode[n_i] != obs_bpm)
                    printlog ("WARNING - GPREPARE: Using an IMAGE mode BPM to flag spectroscopic data!",
		        l_logfile,verbose+)
            }

            # Start the loop over the SCI extensions
            n_j = 1
            next_out = next_inp + 2*next_sci
            while (n_j<=next_sci) {
        	n_var[n_j] = next_inp + n_j
        	n_dq[n_j] = next_inp + next_sci + n_j
        	tmpv[n_j] = mktemp("tmpv")
        	tmpd[n_j] = mktemp("tmpd")
        	tmpdq = mktemp("tmpdq")
        	tmpsci = outimg[n_i]//"["//l_sci_ext//","//n_j//"]"

                # Get RN and GAIN from header or from parameters and calculate 
		# the VAR frame

        	imgets(tmpsci, l_key_gain, >>& "dev$null")
        	if (imgets.value != "0")
                    l_gain = real(imgets.value)
        	else {
                    printlog("WARNING - GPREPARE: Gain not found in image header. Using a gain",
		        l_logfile,verbose+)
                    printlog("                    of "//l_gain//" electrons per ADU.",
		        l_logfile,verbose+)
        	} 

        	imgets(tmpsci,l_key_ron, >>& "dev$null")
        	if (imgets.value != "0")
                    l_ron = real(imgets.value)
        	else {
                    printlog("WARNING - GPREPARE: Read noise not found in image header.",
		        l_logfile,verbose+)
                    printlog("                    Using a read noise of "//l_ron//" electrons.",
		        l_logfile,verbose+)
        	}

                # Do the math
        	imexpr(varexp,tmpv[n_j],l_ron,l_gain,tmpsci,dims="auto",
		    intype="auto",outtype="real",refim="auto",bwidth=0,
		    btype="nearest",bpixval=0.,rangecheck=yes,verbose=no,
		    exprdb="none",lastout="")

                # Create DQ frame and combine with the BPM if one is given
                # Identify saturated pixels
        	imexpr(dqexp,tmpdq,tmpsci,l_sat,dims="auto",intype="auto",
		    outtype="short",refim="auto",bwidth=0,btype="nearest",
		    bpixval=0.,rangecheck+,verbose-,exprdb="none",lastout="",
		    mode="ql")

                # Check if correspond to the same region of the CCD (DATASEC 
		# and TRIMSEC keywords)
                # Needs to be modified to handle windowed readout images.
        	if (l_bpm != "") {              
                    imgets(tmpsci,l_key_dsec, >>& "dev$null")
                    sci_dsec = imgets.value
                    imgets(l_bpm//"["//l_sci_ext//","//n_j//"]",l_key_dsec,
		        >>& "dev$null")
                    bpm_dsec = imgets.value

                    # will crash here if image is in windowed readout mode.

                    if (sci_dsec != bpm_dsec) {
                	printlog("ERROR - GPREPARE: Science and DQ frame for extension ["//l_sci_ext//","//n_j//"] of "//img,
			    l_logfile,yes)
                	printlog("                  correspond to different data sections.",
			    l_logfile,yes)
                	imdelete (outimg[n_i]//",tmpv*,"//tmpdq,ver-, >>& "dev$null")
                	goto crash
                    } 

                    imgets (tmpsci, "TRIMSEC", >>& "dev$null")
                    sci_tsec = imgets.value

                    imgets (l_bpm//"["//l_sci_ext//","//n_j//"]", "TRIMSEC",
		        >>& "dev$null")
                    bpm_tsec = imgets.value

                    if (bpm_tsec == "0" && sci_tsec != "0") {
                	tsec = sci_tsec
                    } else if (sci_tsec == "0" && bpm_tsec != "0") {
                	tsec = bpm_tsec
                    } else if (sci_tsec != "0" && sci_tsec != bpm_tsec) {
                	printlog("ERROR - GPREPARE: Science and DQ frame for extension ["//l_sci_ext//","//n_j//"] of "//img,
			    l_logfile,yes)
                	printlog("                  correspond to different trimming sections.",
			    l_logfile,yes)
                	imdelete (outimg[n_i]//",tmpv*,"//tmpdq,ver-,
			    >>& "dev$null")
                	goto crash
                    } else
                	tsec = ""
			
                    addmasks(tmpdq//tsec//","//l_bpm//"["//l_sci_ext//","//n_j//"]"//tsec,
		        tmpd[n_j]//".fits","im1 || im2")                 
                    if (tsec != "" ) { 
                	date | scan(l_struct)
                	hedit(tmpd[n_j],"TRIMSEC",tsec,add+,addonly-,delete-,
		            verify-,show-,update+)
                    }
        	} else
                    imcopy(tmpdq,tmpd[n_j]//".fits",ver-, >>& "dev$null")      

        	imdelete(tmpdq,ver-, >>& "dev$null")

        	n_j = n_j+1     
            } # end of while(n_j) loop

            # Build the expressions for fxinsert 
            varpos = next_inp
            dqpos = next_inp + next_sci
            fxvexp = tmpv[1]//".fits"
            fxdexp = tmpd[1]//".fits"
            n_j = 2
            while (n_j<=next_sci) {
        	fxvexp = fxvexp//","//tmpv[n_j]//".fits"
        	fxdexp = fxdexp//","//tmpd[n_j]//".fits"   
        	n_j = n_j+1         
            }

            # And insert         
            fxinsert(fxvexp,outimg[n_i]//"["//varpos//"]","",verbose-,
	        >>& "dev$null")
            fxinsert(fxdexp,outimg[n_i]//"["//dqpos//"]","",verbose-,
	        >>& "dev$null")

            # Update the headers and clean up temporaries         
            n_j = 1         
            while(n_j<=next_sci) {         
                hedit(outimg[n_i]//"["//n_var[n_j]//"]","EXTNAME",l_var_ext,add+,
		    addonly-,delete-,verify-,show-,update+)
                hedit(outimg[n_i]//"["//n_var[n_j]//"]","EXTVER",n_j,add+,
		    addonly-,delete-,verify-,show-,update+)
                hedit(outimg[n_i]//"["//n_dq[n_j]//"]","EXTNAME",l_dq_ext,add+,
		    addonly-,delete-,verify-,show-,update+)
                hedit(outimg[n_i]//"["//n_dq[n_j]//"]","EXTVER",n_j,add+,
		    addonly-,delete-,verify-,show-,update+)
                imdelete(tmpv[n_j]//","//tmpd[n_j],ver-, >>& "dev$null")
                n_j = n_j+1
            }

	} #end of if(fl_vardq) loop

        # Now update the PHU
        # get the pixscale, use binning in Y
        ybin=1
        keypar(outimg[n_i]//"["//l_sci_ext//",1]","CCDSUM",silent+)
        if (keypar.found)
	    print(keypar.value) | scan(dum,ybin)
        else
	    printlog("WARNING - GPREPARE: Cannot find CCDSUM, assume PIXSCALE="//pixscale[inst],
	        l_logfile,l_verbose)

	gemhedit(outimg[n_i]//"[0]","PIXSCALE",(pixscale[inst]*ybin),
	    "Pixel scale in Y in arcsec/pixel")

	date | scan(l_struct)
	gemhedit(outimg[n_i]//"[0]","NEXTEND",next_out,"Number of extensions")
	gemhedit(outimg[n_i]//"[0]","OBSMODE",obsmode[n_i],
	    "Observing mode (IMAGE|IFU|MOS|LONGSLIT)")
	gemhedit(outimg[n_i]//"[0]","GPREPARE",l_struct,
	    "Time stamp for GPREPARE")
	gemhedit(outimg[n_i]//"[0]","GEM-TLM",l_struct,
	    "Last modification with GEMINI")    
	if (l_bpm == "") 
            test ="none"
	else 
            test = l_bpm

	if (fl_vardq) 
            gemhedit (outimg[n_i]//"[0]","BPMMASK",test,"Input BPM file")

        # update only if keyword not defined
	if (l_fl_addmdf && addmdf[n_i] && l_key_mdf == "") {
            fparse(mdfname[n_i],verb-)
            gemhedit (outimg[n_i]//"[0]", "MASKNAME",fparse.root,"Input MDF")
	}
	n_i+=1

    } # end of while(i) loop

    goto clean

    #--------------------------------------------------------------------------
crash:
    # Exit with error subroutine
    status=1
    goto clean

clean:
    # clean up
    delete(filelist,ver-, >>& "dev$null")
    scanfile=""

    #close log file
    if (status == 0) {
	printlog("   ",l_logfile,l_verbose)
	printlog("GPREPARE exit status:  good.",l_logfile,l_verbose)
	printlog("----------------------------------------------------------------------------",
	    l_logfile,l_verbose)
    } else {
	printlog("   ",l_logfile,l_verbose)
	printlog("GPREPARE exit status:  error.",l_logfile,l_verbose)
	printlog("----------------------------------------------------------------------------",
	    l_logfile,l_verbose)
    }

end
