# GNIRS data reduction script
#
# This data processing was done for data quality assessment by Gemini staff.
# The processing is not designed to give the best possible result.  The user
# of these data is encouraged to use the provided "final" images, and the
# processing steps described here, only as a guide and to re-reduce the
# data carefully to obtain the best possible reduction.
#
# Observation UT date: 2004 Apr 13
# Data filename prefix: S20040413S
#
# Comments from data processor:
#
# This is an example of a basic calibration reduction of a cross-dispersed/32 
#   (R~2000) dataset.  The data consist of GCAL flats, an arc, a pinhole and 
#   a telluric standard.  The obslog entries included below (just before the 
#   processing steps) give a brief description of the data.
#
#
# This line points to the raw data 
set rawdir=/net/veronica/staging/gnirs/2004apr13/

# Load required tasks/packages
gemini
gnirs

#-----------setup for gnirs-----------
#set default parameters
#use gnirs defaults except where specified on command lines below
unlearn gemini
unlearn gemtools
unlearn gnirs

#define logfile and database
#to start from scratch, delete existing log file
delete GS-CAL20040413.log
gnirs.logfile="GS-CAL20040413.log"
#set database dir
gnirs.database="GS-CAL20040413_database"
#to start from scratch, delete database entries if present
if (access("GS-CAL20040413_database")) delete GS-CAL20040413_database/*

#load default header keywords for gnirs
nsheaders gnirs
set stdimage=imt1024


###Start processing

#--------Obslog entries for SV-11 Calibration Data------ 
#GS-CAL20040413-11 382-389  03:05 HD 97411   abba  XD 1.0/SCXD 32/2.208 SBC/XD 2/1/3      Telluirc for SV-11, AM=1.05; changed exp. time, use 386-389
#GS-CAL20040413-14 451-454  06:27 HD105764   abba  XD 1.0/SCXD 32/2.208 SBC/XD 10/1/2     Telluirc for SV-11, AM=1.4 
#GS-CAL20040413-16 456-480  06:57 GCAL QH&Ar  -    XD 1.0/SCXD 32/2.208 SBC/XD 0.2,10,1/1/1 GCAL Flats (21) and 2 arcs (10sec & 1sec)
#GS-CAL20040413-17 481-504  06:57 GCAL QH     -    XD 1.0/SCXD 32/2.208 SBC/XD 0.2,10,1/1/1 more GCAL Flats using order sorters 5&6 for blue end & pinhole; use 483-491 & 493-502, and 504 for pinholes

#---------Create lists-------------
del *.lis
gemlist S20040413S 386-389 > telluric1.lis
gemlist S20040413S 451-454 > telluric2.lis
gemlist S20040413S "456-464,466-474,476-478" > flat1.lis
gemlist S20040413S 480-480 > arc.lis
gemlist S20040413S "483-491" > flat2.lis
gemlist S20040413S "493,496-502" > flat3.lis
gemlist S20040413S 504-504 > pinhole.lis
cat flat*.lis > allflats.lis
cat telluric*.lis > alltellurics.lis
concat allflats.lis,arc.lis,pinhole.lis,alltellurics.lis > allcal.lis

#-------prepare cals----
# nprepare
imdel n@allcal.lis
nsprepare @allcal.lis rawpath=rawdir$ shiftx=INDEF shifty=INDEF
# shift was 2.8

# reduce pinhole
imdel rn@pinhole.lis
nsreduce n@pinhole.lis fl_nscut+ fl_nsappwave- fl_dark- fl_sky- fl_flat- \
  fl_vardq-

# Now obtain the S-distortion
del GS-CAL20040413_database/idrn*
nssdist rn@pinhole.lis coordlist=gnirs$data/pinholes-short-dense.lis \
  section="[*,100]" fl_inter- function="legendre" order=5 verbose+ \
  aiddebug="l" minsep=5 thresh=1000

# test
#imdel trn@pinhole.lis
#nstransform rn@pinhole.lis sdist=rn@pinhole.lis fl_inter-
#nxdisplay trnS20040413S0504 1

# Process and trace the reference arc
imdel rn@arc.lis
nsreduce n@arc.lis fl_nscut+ fl_sky- fl_flat-

# wavelength calibration (needs sdist trace for cd data)
del GS-CAL20040413_database/idw*
imdel wrn@arc.lis
nswave rn@arc.lis sdist=rnS20040413S0504 nlost=30
# (no soln for order 6)

# ------generate flatfields--------
imdel rn@allflats.lis
nsreduce n@allflats.lis fl_sky- fl_flat- fl_nsappwave- fl_corner+

#check flats
!echo "--------------------------------------------" >> GS-CAL20040413.log
del tmpflat
gemextn n@flat1.lis proc="expand" extname="SCI" > tmpflat
!echo "Band 3-5 Flats: 456-478" >> GS-CAL20040413.log
imstat @tmpflat >> GS-CAL20040413.log
del tmpflat
gemextn n@flat2.lis proc="expand" extname="SCI" > tmpflat
!echo "Band 6 flats: 483-491" >> GS-CAL20040413.log
imstat @tmpflat >> GS-CAL20040413.log
!echo "--------------------------------------------" >> GS-CAL20040413.log

# Run interactively or use info in apertures.fits
# bands 3-5
del rnS20040413S0456_flat*.*
nsflat rn@flat1.lis trace="" fl_inter- fl_vardq+ order=60 fl_corner+ \
  rejtype="minmax"
#nxdisplay rnS20040413S0456_flat 1
# band 6
del rnS20040413S0483_flat*.*
nsflat rn@flat2.lis trace="" fl_inter- fl_range- fl_vardq+ order=60 \
  fl_corner+ rejtype="minmax"
#nxdisplay rnS20040413S0483_flat 1

# the apertures for the flat are: (line 205)
# [SCI,1] = -26, +16
# [SCI,2] = -26, +16
# [SCI,3] = -26, +15
# [SCI,4] = -27, +12
# [SCI,5] = -26, +17 (using flat2.lis flats)
# [SCI,6] = guess... (using flat2.lis flats)

# create the final flat: order: 1-4 flat1 -> position 1-12, 
#                               5-6 flat2 -> position 13-18
imdel temp_flat.fits
fxcopy rnS20040413S0456_flat.fits temp_flat.fits "0-12" new_file+
fxinsert rnS20040413S0483_flat.fits temp_flat.fits[12] groups="13-18"
imdel rnS20040413S0456_flat.fits
fxcopy temp_flat.fits rnS20040413S0456_flat.fits "0-18" new_file+
imdel temp_flat.fits,rnS20040413S0483_flat.fits
#nxdisplay rnS20040413S0456_flat

# Standard:

# determinating the time inbetween exposures:
!echo "--------------------------------------------" >> GS-CAL20040413.log
del tmpphu
gemext n@telluric1.lis proc="expand" index="0" > tmpphu
!echo "Telluric 1: 386-389" >> GS-CAL20040413.log
hselect @tmpphu UT yes >> GS-CAL20040413.log
#time between exposures ~14 seconds
del tmpphu
gemext n@telluric2.lis proc="expand" index="0" > tmpphu
!echo "Telluric 1: 451-454" >> GS-CAL20040413.log
hselect @tmpphu UT yes >> GS-CAL20040413.log
del tmpphu
#time between exposures ~27 seconds
!echo "--------------------------------------------" >> GS-CAL20040413.log

# reduce tellurics
# nod size very small...
imdel rn@telluric1.lis,rn@telluric2.lis
nsreduce n@telluric1.lis fl_nscut+ fl_corner+ fl_process+ fl_nsapp- \
  fl_sky+ skyrange=17 fl_flat+ flatimage="rnS20040413S0456_flat.fits" \
  nodsize=1.5
nsreduce n@telluric2.lis fl_nscut+ fl_corner+ fl_process+ fl_nsapp- \
  fl_sky+ skyrange=30 fl_flat+ flatimage="rnS20040413S0456_flat.fits" \
  nodsize=1.5

# Combine, wavelength calibrate and extract
imdel rnS20040413S0386_comb.fits,rnS20040413S0451_comb.fits
nscombine rn@telluric1.lis
nscombine rn@telluric2.lis

# Star is bright enough so we can trace without transforming 
# in the spatial direction
imdel trnS20040413S0386_comb.fits,trnS20040413S0451_comb.fits
nstransform rnS20040413S0386_comb.fits yorder=4 fl_inter- \
  lamptrans=wrnS20040413S0480
nstransform rnS20040413S0451_comb.fits yorder=4 fl_inter- \
  lamptrans=wrnS20040413S0480

# Extract
imdel xtrnS20040413S0386_comb.fits,xtrnS20040413S0451_comb.fits
nsextract trnS20040413S0386_comb.fits line=400 nsum=20 upper=6 low=-6 \
  fl_inter- fl_apall+ fl_trace+
nsextract trnS20040413S0451_comb.fits line=400 nsum=20 upper=6 low=-6 \
  fl_inter- fl_apall+ fl_trace+

#Cleanup: 
#keep only .cl, .log, database, *flat*, wrn*fits (arcs), trn*fits
#(2-D reduced images),
#xtrn*fits (1-D extractions) and *add.fits (final 1-D spectra)
#and pinhole data
imdel n@allcal.lis,rn@allflats.lis,rn@arc.lis,rn@alltellurics.lis
delete tmp*,*.lis



