# Copyright(c) 2004 Association of Universities for Research in Astronomy, Inc.

# acooke 28 Mar 2004

# Generate an aquisition image from IFU data and the table generated
# by nfmake


procedure nfquick (inimages)

char	inimages	{"", prompt = "Data to convert to aq. images"}
char	outimages	{"", prompt = "Output files"}
char	outprefix	{"a", prompt = "Prefix to use if outimages not given"}

char	table		{"gnirs$data/ifu-map.fits", prompt = "Table from nfmake that defines mapping"}
real	shiftx		{0.0, prompt = "Shift in X"}
real	shifty		{0.0, prompt = "Shift in Y"}
real	xscale		{0.15, prompt = "Size of pixels in X (arcec)"}
real	yscale		{0.15, prompt = "Size of pixels in X (arcec)"}
bool	fl_inter	{yes, prompt = "Get shift interactively"}
bool	fl_display	{yes, prompt = "Display each image as it is generated"}
bool	fl_shift	{yes, prompt = "Get shift to centre star in each image"}

char	logfile		{"", prompt = "Logfile"}
bool	verbose		{yes, prompt = "Verbose output?"}

int	status		{0, prompt = "O: Exit status (0 = good)"}
struct*	scanin1		{"", prompt = "Internal use only"}
struct*	scanin2		{"", prompt = "Internal use only"}

begin
	char	l_inimages = ""
	char	l_outimages = ""
	char	l_outprefix = ""
	char	l_table = ""
	real	l_shiftx, l_shifty, l_xscale, l_yscale
	char	l_logfile = ""
	bool	l_fl_inter
	bool	l_fl_display
	bool	l_fl_shift
	bool	l_verbose

	int	junk, nfiles, nx, ny
	bool	debug, compress, havex, havey, havexy
	struct	sdate, sline
	char	tmpin, tmpout, tmplist, inimg, outimg, tmpdump, src, dst
	char	tmproot, tmpwhole, tmpthin, tmp1, tmp2, tmp3, tmp4, tmplog
	real	x1, y1, dx, dy, x2, y2, dy2

	junk = fscan (	inimages, l_inimages)
	junk = fscan (	outimages, l_outimages)
	junk = fscan (	outprefix, l_outprefix)
	junk = fscan (	table, l_table)
	l_shiftx =	shiftx
	l_shifty = 	shifty
	l_xscale =	xscale
	l_yscale =	yscale
	junk = fscan (	logfile, l_logfile)
	l_fl_inter =	fl_inter
	l_fl_display =	fl_display
	l_fl_shift =	fl_shift
	l_verbose =	verbose

	status = 1
	debug = no


	tmpin = mktemp ("tmpin")
	tmpout = mktemp ("tmpout")
	tmplist = mktemp ("tmplist")
	tmpdump = mktemp ("tmpdump")
	tmproot = mktemp ("tmproot")
	tmplog = mktemp ("tmplog")


	cache ("gemextn", "keypar")


	# Start logging

	if ("" == l_logfile) {
	    junk = fscan (gnirs.logfile, l_logfile) 
	    if (l_logfile == "") {
		l_logfile = "gnirs.log"
		printlog ("WARNING - NFQUICK: Both gnswedit.logfile and \
		    gnirs.logfile are empty.", l_logfile, verbose+) 
		printlog ("                   Using default file " \
		    // l_logfile // ".", l_logfile, verbose+) 
	    }
	}
	date | scan (sdate) 
	printlog ("------------------------------------------------------------------------------", 
	    l_logfile, verbose = l_verbose) 
	printlog ("NFQUICK -- " // sdate, l_logfile, l_verbose) 
	printlog (" ", l_logfile, l_verbose) 


	# Check input data

	if (debug) print ("checking input")
	gemextn (l_inimages, proc="none", check="", index="", extname="", 
	    extver="", ikparams="", omit="path", replace="", outfile=tmproot, 
	    logfile="", glogpars="", verbose-)
	if (0 < gemextn.fail_count || 0 == gemextn.count) {
	    printlog ("ERROR - NFQUICK: Missing input data.", \
		l_logfile, verbose+) 
	    goto clean
	}
	nfiles = gemextn.count
	gemextn (l_inimages, proc="expand", check="exist,mef", index="1", \
	    extname="", extver="", ikparams="", omit="", replace="", \
	    outfile=tmpin, logfile="", glogpars="", verbose-)
	if (0 < gemextn.fail_count || nfiles != gemextn.count) {
	    printlog ("ERROR - NFQUICK: Bad input data.", l_logfile, verbose+) 
	    goto clean
	}


	# Check/generate output names

	if (debug) print ("checking output")
	gemextn (l_outimages, proc="none", check="absent", index="", \
	    extname="", extver="", ikparams="", omit="", replace="", \
	    outfile=tmpout, logfile="", glogpars="", verbose-)
	if (0 < gemextn.fail_count) {
	    printlog ("ERROR - NFQUICK: Output data already exist.", \
		l_logfile, verbose+) 
	    goto clean
	}
	if (0 != gemextn.count) {
	    if (nfiles != gemextn.count) {
		printlog ("ERROR - NFQUICK: Incorrect number of output \
		    files.", l_logfile, verbose+) 
		goto clean
	    }
	} else {
	    delete (tmpout, ver-, >& "dev$null")
	    gemextn ("%^%" // l_outprefix // "%" // "@" // tmproot,
		proc="none", check="absent", index="", extname="", \
		extver="", ikparams="", omit="kernel", replace="", \
		outfile=tmpout, logfile="", glogpars="", verbose-)
	    if (0 < gemextn.fail_count || nfiles != gemextn.count) {
		printlog ("ERROR - NFQUICK: Bad output files.", \
		    l_logfile, verbose+) 
		goto clean
	    }
	}


	# Combine in + out into single list

	join (tmpin, tmpout, output = tmplist, delim = " ", \
	    missing = "Missing", maxchar = 161, shortest = yes, \
	    verbose-)


	# Get output image size

	if (debug) print ("reading table")
	if (no == access (l_table)) {
	    printlog ("ERROR - NFQUICK: Cannot access table " // l_table \
		// ".", l_logfile, verbose+) 
	    goto clean
	}
	keypar (l_table, "NFQ_X", silent+)
	if (keypar.found) {
	    nx = int (keypar.value)
	} else {
	    printlog ("ERROR - NFQUICK: Bad table.", l_logfile, verbose+) 
	    goto clean
	}

	keypar (l_table, "NFQ_Y", silent+)
	if (keypar.found) {
	    ny = int (keypar.value)
	} else {
	    printlog ("ERROR - NFQUICK: Bad table.", l_logfile, verbose+) 
	    goto clean
	}


	# Get centring interactively, if required

	if (l_fl_inter) {

	    havex = no; havey = no
	    while (! havex || ! havey) {

		if (no == havey)
		    printlog ("NFQUICK: press k to identify the slit \
			in the y direction", l_logfile, verbose+) 
		if (no == havex)
		    printlog ("NFQUICK: press j to identify a central gap \
			in the x direction", l_logfile, verbose+)

		scanin1 = tmplist
		junk = fscan (scanin1, inimg)
		imexam (inimg, frame=1, image="", logfile=tmplog, keeplog+)

		if (debug) type (tmplog)

		scanin2 = tmplog
		while (fscan (scanin2, sline) != EOF) {
		    if (1 == stridx (sline, "Column")) {
			havey = yes
			junk = fscan (sline, tmp1, tmp2, tmp3, tmp4)
			if (strlen (tmp3) <= 8) sline = tmp4
			else sline = substr (tmp3, 8, strlen (tmp3))
			junk = fscan (sline, l_shifty)
			l_shifty = 511 - l_shifty
		    }
		    if (1 == stridx (sline, "Line")) {
			havex = yes
			junk = fscan (sline, tmp1, tmp2, tmp3, tmp4)
			if (strlen (tmp3) <= 8) sline = tmp4
			else sline = substr (tmp3, 8, strlen (tmp3))
			junk = fscan (sline, l_shiftx)
		    }
		}

		if (debug) print ("trying to delete log")
		delete (tmplog, ver-, >& "dev$null")
	    }
	}


	# Generate the images

	scanin1 = tmplist

	while (fscan (scanin1, inimg, outimg) != EOF) {

	    if (debug)
		print ("mkimage " // outimg // "[" // nx // "," // ny // "]")

	    mkimage (outimg, option = "make", value = 0.0, ndim = 2, \
		dims = nx // " " // ny, pixtype = "real", slope = 0.0, \
		sigma = 0.0, seed = 0)

	    delete (tmpdump, ver-, >& "dev$null")

	    tprint (l_table, prparam-, prdata+, pwidth = 1000, \
		plength = INDEF, showrow-, orig_row-, showhdr-, \
		columns = "x1,y1,dx,dy,x2,y2,compress", rows = "-", \
		option = "plain", align-, sp_col = "", lgroup = 0, \
		>> tmpdump)


	    # Get nearest appropriate x shift

	    if (l_fl_inter) {

		scanin2 = tmpdump
		havex = no
		while (! havex &&
		    fscan (scanin2, x1, y1, dx, dy, x2, y2, compress) != EOF) {

		    if (abs (x1 - l_shiftx) < 0.5 * dx) {
			if (debug) print (x1 // ", " // l_shiftx)
			l_shiftx = x1 - l_shiftx
			havex = yes
		    }
		}
		
		if (no == havex) {
		    printlog ("ERROR - NFQUICK: Cannot find an x offset that \
			matches " // l_shiftx, l_logfile, verbose+) 
		    goto clean
		}
		
		printlog ("NFQUICK: Offsets: x: " // l_shiftx // "; y: " \
		    // l_shifty, l_logfile, l_verbose) 
	    }


	    scanin2 = tmpdump
	    if (debug) type (tmpdump)

	    while (fscan (scanin2, x1, y1, dx, dy, x2, y2, compress) != EOF) {

		src = "[" // int (x1-l_shiftx) // ":" // int (x1+dx-l_shiftx) \
		    // "," \
		    // int (y1-l_shifty) // ":" // int (y1+dy-l_shifty) // "]"
		src = inimg // src

		if (compress) {
		    dy2 = 1
		    tmpwhole = mktemp ("tmpwhole")
		    tmpthin = mktemp ("tmpthin")
		    imcopy (src, tmpwhole, >& "dev$null")
		    improject (tmpwhole, tmpthin, 2, >& "dev$null")
		    src = tmpthin
		} else {
		    dy2 = dy
		}

		dst = "[" // int (x2) // ":" // int (x2+dx) \
		    // "," \
		    // int (y2) // ":" // int (y2+dy2) // "]"

		if (debug) print (src // " -> " // outimg // dst)

		imcopy (src, outimg // dst, >& "dev$null")

		imdel (tmpwhole, ver-, >& "dev$null")
		imdel (tmpthin, ver-, >& "dev$null")

	    }

	    # Give shift info for star centre

	    if (l_fl_shift) {

		havexy = no
		while (! havexy) {

		    printlog ("NFQUICK: press a or x to identify star \
			and then q to quit", l_logfile, verbose+) 

		    imexam (outimg, frame=2, image="", logfile=tmplog, \
			keeplog+)

		    if (debug) type (tmplog)

		    scanin2 = tmplog
		    x1 = -1.0
		    y1 = -1.0
		    while (! havexy && fscan (scanin2, x1, y1) != EOF) {
			if (x1 >= 0.0 && y1 >= 0.0) {
			    havexy = yes
			} else {
			    x1 = -1.0
			    y1 = -1.0
			}				
		    }

		    if (debug) print ("trying to delete log")
		    delete (tmplog, ver-, >& "dev$null")

		    x2 = 0.5 * nx
		    y2 = 0.5 * ny
		    dx = x1 - x2
		    dy = y1 - y2
		    printlog (" ", l_logfile, verbose+)
		    printf ("%16s: star (%5.2f, %5.2f); image [%2d, %2d]; \
			shift (%5.2f, %5.2f);\n", outimg, x1, y1, nx, ny, \
			dx, dy) | scan (sline)
		    printlog (sline, l_logfile, verbose+)
		    printf ("                  arcsec (%5.2f, %5.2f)\n", \
			dx * l_xscale, dy * l_yscale) | scan (sline)
		    printlog (sline, l_logfile, verbose+)
		}

	    } else if (l_fl_display) {
		display (outimg, 2)
	    }
	    
	}



	status = 0	# Success

clean:

	scanin1 = ""
	scanin2 = ""

	delete (tmpin, ver-, >& "dev$null")
	delete (tmpout, ver-, >& "dev$null")
	delete (tmplist, ver-, >& "dev$null")
	delete (tmpdump, ver-, >& "dev$null")
	delete (tmproot, ver-, >& "dev$null")

	if (0 == status) {
	    printlog (" ", l_logfile, l_verbose) 
	    printlog ("NFQUICK exit status:  good.", l_logfile, l_verbose) 
	}
	
end
