# Copyright(c) 2004 Association of Universities for Research in Astronomy, Inc.

# acooke 7 Apr 2004

# measure the x shift in ifu data (for nsprepare.shiftx)


procedure nfxshift (inimage)

char	inimage		{"", prompt = "Image to measure"}

bool	fl_inter	{yes, prompt = "Get shift interactively"}

char	configtable	{"gnirs$data/config.fits", prompt = "Table of configuration settings"}
char	logfile		{"", prompt = "Logfile"}
bool	verbose		{yes, prompt = "Verbose output?"}

real	shiftx		{0.0, prompt = "Output - shift in X"}
int	status		{0, prompt = "O: Exit status (0 = good)"}
struct*	scanin1		{"", prompt = "Internal use only"}

begin
	char	l_inimage = ""
	char	l_logfile = ""
	bool	l_fl_inter
	bool	l_verbose
	char	l_key_decker = ""
	char	l_key_grating = ""
	char	l_key_prism = ""
	char	l_configtable = ""
	real	l_shiftx

	int	junk
	bool	debug, havex, first
	struct	sdate, sline
	char	badhdr, phu, prism, decker, grating, mdf
	char	tmpin, tmpphu, inimg, tmpdump, src, dst
	char	tmp1, tmp2, tmp3, tmp4, tmplog, tmpmdfrow
	real	before, x

	junk = fscan (	inimage, l_inimage)
	junk = fscan (	configtable, l_configtable)
	junk = fscan (	logfile, l_logfile)
	l_fl_inter =	fl_inter
	l_verbose =	verbose

	badhdr = ""
	junk = fscan (	nsheaders.key_decker, l_key_decker)
	if ("" == l_key_decker) badhdr = badhdr + " key_decker"
	junk = fscan (	nsheaders.key_grating, l_key_grating)
	if ("" == l_key_grating) badhdr = badhdr + " key_grating"
	junk = fscan (	nsheaders.key_prism, l_key_prism)
	if ("" == l_key_prism) badhdr = badhdr + " key_prism"

	status = 1
	debug = no


	tmpin = mktemp ("tmpin")
	tmpphu = mktemp ("tmpphu")
	tmplog = mktemp ("tmplog")
	tmpdump = mktemp ("tmpdump")
        tmpmdfrow = mktemp ("tmpmdfrow") 


	cache ("gemextn", "keypar")


	# Start logging

	if ("" == l_logfile) {
	    junk = fscan (gnirs.logfile, l_logfile) 
	    if (l_logfile == "") {
		l_logfile = "gnirs.log"
		printlog ("WARNING - NFXSHIFT: Both gnswedit.logfile and \
		    gnirs.logfile are empty.", l_logfile, verbose+) 
		printlog ("                   Using default file " \
		    // l_logfile // ".", l_logfile, verbose+) 
	    }
	}
	date | scan (sdate) 
	printlog ("------------------------------------------------------------------------------", 
	    l_logfile, verbose = l_verbose) 
	printlog ("NFXSHIFT -- " // sdate, l_logfile, l_verbose) 
	printlog (" ", l_logfile, l_verbose) 


	if ("" != badhdr) {
            printlog ("ERROR - NFXSHIFT: Parameter(s) missing from \
		nsheaders: " // badhdr, l_logfile, verbose+) 
            goto clean
	}
                

	# Check input data

	if (debug) print ("checking input")
	gemextn (l_inimage, proc="expand", check="exist,mef", index="1", \
	    extname="", extver="", ikparams="", omit="", replace="", \
	    outfile=tmpin, logfile="", glogpars="", verbose-)
	if (0 < gemextn.fail_count || 1 != gemextn.count) {
	    printlog ("ERROR - NFQUICK: Bad input data.", l_logfile, verbose+) 
	    goto clean
	}

	scanin1 = tmpin
	junk = fscan (scanin1, inimg)
	
	gemextn (l_inimage, proc="expand", check="exist,mef", index="0", \
	    extname="", extver="", ikparams="", omit="", replace="", \
	    outfile=tmpphu, logfile="", glogpars="", verbose-)
	if (0 < gemextn.fail_count || 1 != gemextn.count) {
	    printlog ("ERROR - NFQUICK: Bad input data.", l_logfile, verbose+) 
	    goto clean
	}

	scanin1 = tmpphu
	junk = fscan (scanin1, phu)


	# Header info needed for MDF

	keypar (phu, l_key_prism, silent+) 
	if (no == keypar.found) {
	    printlog ("ERROR - NFXSHIFT: Could not read prism \
		for " // inimg, l_logfile, verbose+) 
	    goto clean
	} else {
	    prism = keypar.value
	}

	keypar (phu, l_key_decker, silent+) 
	if (no == keypar.found) {
	    printlog ("ERROR - NFXSHIFT: Could not read decker \
		for " // inimg, l_logfile, verbose+) 
	    goto clean
	} else {
	    decker = keypar.value
	}

	keypar (phu, l_key_grating, silent+) 
	if (no == keypar.found) {
	    printlog ("ERROR - NFXSHIFT: Could not read grating \
		for " // inimg, l_logfile, verbose+) 
	    goto clean
	} else {
	    grating = keypar.value
	}


	# Get centring interactively, if required
	# (currently no non-interaactive implementation)

	if (l_fl_inter) {

	    havex = no
	    while (! havex) {

		if (no == havex)
		    printlog ("NSHIFTX: press j to identify a central gap \
			in the x direction", l_logfile, verbose+)

		imexam (inimg, logfile=tmplog, keeplog+)

		if (debug) type (tmplog)

		scanin1 = tmplog
		while (fscan (scanin1, sline) != EOF) {
		    if (1 == stridx (sline, "Line")) {
			havex = yes
			junk = fscan (sline, tmp1, tmp2, tmp3, tmp4)
			if (strlen (tmp3) <= 8) sline = tmp4
			else sline = substr (tmp3, 8, strlen (tmp3))
			junk = fscan (sline, l_shiftx)
		    }
		}

		if (debug) print ("trying to delete log")
		delete (tmplog, ver-, >& "dev$null")
	    }
	}


	# Get nearest appropriate x shift

	tselect (intable = l_configtable, outtable = tmpmdfrow, \
	    expr = "prism='" // prism // "'&&decker='" \
	    // decker // "'&&grating='" // grating // "'")
	tinfo (table = tmpmdfrow, >>& "dev$null")
	if (0 != tinfo.nrows) {
	    tabpar (table = tmpmdfrow, column = "mdf", row = 1, \
		format-)
	    mdf = tabpar.value
	} else {
	    printlog ("ERROR - NFXSHIFT: Could not read MDF.", 
		l_logfile, verbose+)
	    printf ("                  '%s' '%s' '%s'\n", prism, decker, \
		grating) | scan (sline)
	    printlog (sline, l_logfile, verbose+)
	    goto clean
	}


	tprint (mdf, prparam-, prdata+, pwidth = 1000, \
	    plength = INDEF, showrow-, orig_row-, showhdr-, \
	    columns = "x_ccd", rows = "-", \
	    option = "plain", align-, sp_col = "", lgroup = 0, \
	    >> tmpdump)


	scanin1 = tmpdump
	havex = no
	first = yes
	while (! havex && fscan (scanin1, x) != EOF) {

	    if (first) {
		first = no
	    } else {
		if (before < l_shiftx && l_shiftx < x) {
		    havex = yes
		    l_shiftx = (before+x)/2.0 - l_shiftx
		}
	    }

	    before = x
	}

	if (no == havex) {
	    printlog ("ERROR - NFXSHIFT: Cannot find an x offset that \
		matches " // l_shiftx, l_logfile, verbose+) 
	    goto clean
	}

	printlog ("NFXSHIFT: Offset in x: " // l_shiftx, l_logfile, l_verbose) 


	shiftx = l_shiftx
	status = 0	# Success

clean:

	scanin1 = ""

	delete (tmpin, ver-, >& "dev$null")
	delete (tmpphu, ver-, >& "dev$null")
	delete (tmpmdfrow//".tab", ver-)#, >& "dev$null")
	delete (tmpdump, ver-, >& "dev$null")

	if (0 == status) {
	    printlog (" ", l_logfile, l_verbose) 
	    printlog ("NFXSHIFT exit status:  good.", l_logfile, l_verbose) 
	}
	
end
