# Copyright(c) 2004 Association of Universities for Research in Astronomy, Inc.

# acooke 27 aug 2004

# measure the shift of one image relative to another
# *plus* cool (trez cool, mon cherry) extra features like accepting 
# mdfs and updating wcs values

procedure nsoffset (refimage, inimages)

char	refimage	{"", prompt = "Reference image"}
char	inimages	{"", prompt = "Images to shift"}
char	outimages	{"", prompt = "Output files"}
char	outprefix	{"o", prompt = "Prefix to use if outimages not given"}
real	shift		{0, prompt = "Measured/required shift"}
int	axis		{INDEF, prompt = "Shift axis"}
real	pixscale	{INDEF, prompt = "Pixel scale (if not in header)"}
real	sigma		{5, prompt = "Initial peak width"}
bool	fl_measure	{yes, prompt = "Measure shift?"}
bool	fl_apply	{no, prompt = "Apply shift to input images?"}
bool	fl_all		{yes, prompt = "Measure all input images?"}
char	logfile		{"", prompt = "Logfile"}
bool	fl_inter	{no, prompt = "Find peak interactively?"}
bool	verbose		{yes, prompt = "Verbose output?"}

int	status		{0, prompt = "Exit status (0 = good)"}
struct*	scanin1		{"", prompt = "Internal use only"}
struct*	scanin2		{"", prompt = "Internal use only"}

begin
	char	l_refimage = ""
	char	l_inimages = ""
	char	l_outimages = ""
	char	l_outprefix = ""
	real	l_shift
	int	l_axis = ""
	real	l_pixscale
	real	l_sigma
	bool	l_fl_measure
	bool	l_fl_apply
	bool	l_fl_all
	char	l_logfile = ""
	bool	l_fl_inter
	bool	l_verbose

	char	l_sci_ext = ""
	char	l_dq_ext = ""
        char	l_key_dispaxis = ""
        char	l_key_pixscale = ""

	int	junk, dispaxis, nfiles, nsecns, isecn, specorder, slitid
	int	nx, ny, fnx, fny
	int 	xc, yc
	bool	debug, first, havemdf
	struct	sline
	char	badhdr, refsci, imgin, imgout, imgsci, mdfin, refphu
	char	slittype, target, defkey, sjunk, mdfout, imgdq, refdq
	char	tmpref, tmprefsci, tmproot, tmpout, tmplist, tmpin
	char	tmpmdfimg, tmparith, tmptapref, tmptaptag, tmpconv
	char	tmpcursor, tmplog, tmptab, tmpdq, tmpblank, tmpmask
	real	scale, x1, x2, y1, y2, corner, x_ccd, y_ccd
	real	slitpos_mx, slitpos_my, slitsize_mx, slitsize_my
	real	slittilit_m, slitsize_mr, slitsize_mw, cmin, cmax
	real	crpix, border, blur


	junk = fscan (	refimage, l_refimage)
	junk = fscan (	inimages, l_inimages)
	junk = fscan (	outimages, l_outimages)
	junk = fscan (	outprefix, l_outprefix)
	l_shift =	0.1 * nint (10 * shift)	# restrict to 0.1 resoln
	l_axis =	axis
	l_pixscale =	pixscale
	l_sigma =	sigma
	l_fl_measure =	fl_measure
	l_fl_apply =	fl_apply
	l_fl_all =	fl_all
	junk = fscan (	logfile, l_logfile)
	l_fl_inter =	fl_inter
	l_verbose =	verbose

	status = 1
	scale = INDEF
	debug = no
	badhdr = ""
	nx = INDEF
	ny = INDEF
	if (l_fl_measure) l_shift = INDEF

	# these chosen for ifu mdf, but should work with anything larger
	blur = 2
	border = -5

        junk = fscan (nsheaders.sci_ext, l_sci_ext)
	if ("" == l_sci_ext) badhdr = badhdr + " sci_ext"
        junk = fscan (nsheaders.dq_ext, l_dq_ext)
	if ("" == l_dq_ext) badhdr = badhdr + " dq_ext"
        junk = fscan (nsheaders.key_dispaxis, l_key_dispaxis)
	if ("" == l_key_dispaxis) badhdr = badhdr + " key_dispaxis"
        junk = fscan (nsheaders.key_pixscale, l_key_pixscale)
	if ("" == l_key_pixscale) badhdr = badhdr + " key_pixscale"

	cache ("gemextn", "minmax")

	tmpref =	mktemp ("tmpref")
	tmprefsci =	mktemp ("tmprefsci")
	tmproot =	mktemp ("tmproot")
	tmpout =	mktemp ("tmpout")
	tmplist =	mktemp ("tmplist")
	tmpin =		mktemp ("tmpin")
	tmpdq =		mktemp ("tmpdq")
	tmpblank =	mktemp ("tmpblank")
	tmpmask =	mktemp ("tmpmask")
	tmpmdfimg =	mktemp ("tmpmdfimg")
	tmparith =	mktemp ("tmparith")
	tmptapref =	mktemp ("tmptapref")
	tmptaptag =	mktemp ("tmptaptag")
	tmpconv =	mktemp ("tmpconv")
	tmpcursor =	mktemp ("tmpcursor")
	tmplog =	mktemp ("tmplog")
	tmptab =	mktemp ("tmptab")

	if ("" == l_logfile) {
	    junk = fscan (gnirs.logfile, l_logfile) 
	    if (l_logfile == "") {
		l_logfile = "gnirs.log"
		printlog ("WARNING - NSOFFSET: Both nsoffset.logfile and \
		    gnirs.logfile are empty.", l_logfile, verbose+) 
		printlog ("                   Using default file " \
		    // l_logfile // ".", l_logfile, verbose+) 
	    }
	}

	date | scan (sline) 
	printlog ("---------------------------------------------------------\
	    ---------------------", l_logfile, verbose = l_verbose) 
	printlog ("NSOFFSET -- " // sline, l_logfile, l_verbose) 
	printlog (" ", l_logfile, l_verbose) 


	if ("" != badhdr) {
            printlog ("ERROR - NSOFFSET: Parameter(s) missing from \
		nsheaders: " // badhdr, l_logfile, verbose+) 
            goto clean
	}


	if (debug) print ("checking logic")
	if (no == l_fl_measure && no == l_fl_apply) {
	    printlog ("ERROR - NSOFFSET: Neither measuring nor applying \
		a shift.", l_logfile, verbose+) 
	    goto clean
	}
	if (no == l_fl_measure && l_fl_all) {
	    printlog ("ERROR - NSOFFSET: Cannot measure all shifts if \
		measuring none", l_logfile, verbose+) 
	    printlog ("                  (fl_all+ but fl_measure-).", \
		l_logfile, verbose+) 
	    goto clean
	}

	if (debug) print ("checking reference")
	gemextn (l_refimage, proc="none", check="exists", index="", \
	    extname="", extver="", ikparams="", omit="", replace="", \
	    outfile=tmpref, logfile="", glogpars="", verbose-)
	if (0 < gemextn.fail_count || 1 != gemextn.count) {
	    printlog ("ERROR - NSOFFSET: Error in reference image.", \
		l_logfile, verbose+) 
	    goto clean
	}
	scanin1 = tmpref; junk = fscan (scanin1, l_refimage)
	refphu = l_refimage // "[0]"

	if (debug) print ("getting first sci extension from ref")
	gemextn (l_refimage, proc="expand", check="exists", index="", \
	    extname=l_sci_ext, extver="1-", ikparams="", omit="", replace="", \
	    outfile=tmprefsci, logfile="", glogpars="", verbose-)
	if (0 < gemextn.fail_count || 0 == gemextn.count) {
	    printlog ("ERROR - NSOFFSET: Missing data in reference.", \
		l_logfile, verbose+) 
	    goto clean
	}
	scanin1 = tmprefsci; junk = fscan (scanin1, refsci)

	if (debug) print ("getting size of reference image")
	hselect (refsci, "i_naxis1", yes) | scan (nx)
	hselect (refsci, "i_naxis2", yes) | scan (ny)
	if (isindef (nx) || isindef (ny)) {
	    printlog ("ERROR - NSOFFSET: Cannot read size of " // refsci \
		// ".", l_logfile, verbose+) 
	    goto clean
	}
	if (debug) print (nx // "x" // ny)

	if (debug) print ("zeroing bad pixels")
	delete (tmpdq, verify-, >& "dev$null")
	tmpdq = mktemp ("tmpdq")
	gemextn (l_refimage, proc="expand", check="exists", index="", \
	    extname=l_dq_ext, extver="1-", ikparams="", omit="", replace="", 
	    outfile=tmpdq, logfile="dev$null", glogpars="", verbose-)
	if (0 == gemextn.fail_count && 0 != gemextn.count) {
	    scanin2 = tmpdq
	    junk = fscan (scanin2, refdq)
	    imexpr ("a>0?0:b", tmpmask, refdq, refsci, \
		dims="auto", intype="auto", outtype="auto", \
		refim="auto", bwidth=0, btype="nearest", \
		bpixval=0, rangecheck+, verbose=debug)
	    refsci = tmpmask
	}

	if (debug) print ("clipping at zero")
	imdel (tmparith, verify-, >& "dev$null")
	tmparith = mktemp ("tmparith")
	imexpr ("max(0,a)", tmparith, refsci, \
	    dims="auto", intype="auto", outtype="auto", \
	    refim="auto", bwidth=0, btype="nearest", \
	    bpixval=0, rangecheck+, verbose=debug)
	imdelete (tmpmask, verify-, >& "dev$null")
	tmpmask = tmparith; tmparith = ""; refsci = tmpmask

	if (debug) print ("tapering reference")
	taper (refsci, tmptapref, width="10 %", subtract="none", \
	    function="cosbell", verbose-, >& "dev$null")
	if (debug) display (tmptapref, 1, zsc-, zr+)

	if (debug) print ("checking input")
	gemextn (l_inimages, proc="none", check="exists", index="", \
	    extname="", extver="", ikparams="", omit="exten,kernel,index", \
	    replace="", outfile=tmproot, logfile="", glogpars="", verbose-)
	if (0 < gemextn.fail_count || 0 == gemextn.count) {
	    printlog ("ERROR - NSOFFSET: Missing input data.", \
		l_logfile, verbose+) 
	    goto clean
	}
	nfiles = gemextn.count

	if (debug) print ("getting axis")
	if (isindef (l_axis)) {
	    dispaxis = INDEF
	    hselect (refphu, l_key_dispaxis, yes) | scan (dispaxis)
	    if (isindef (dispaxis)) {
		printlog ("ERROR - NSOFFSET: No axis parameter value and \
		    no " // l_key_dispaxis, l_logfile, verbose+) 
		printlog ("                  in " // refphu, \
		    l_logfile, verbose+) 
		goto clean
	    }
	    l_axis = 3 - dispaxis
	}
	if (debug) print ("axis: " // l_axis)


	if (l_fl_apply) {

	    if (debug) print ("checking output")
	    gemextn (l_outimages, proc="none", check="absent", index="", \
		extname="", extver="", ikparams="", omit="", replace="", \
		outfile=tmpout, logfile="", glogpars="", verbose-)
	    if (0 < gemextn.fail_count) {
		printlog ("ERROR - NSOFFSET: Output data already exist.", \
		    l_logfile, verbose+) 
		goto clean
	    }
	    if (0 != gemextn.count) {
		if (nfiles != gemextn.count) {
		    printlog ("ERROR - NSOFFSET: Incorrect number of output \
			files.", l_logfile, verbose+) 
		    status = 2
		    goto clean
		}
	    } else {
		delete (tmpout, ver-, >& "dev$null")
		gemextn ("%^%" // l_outprefix // "%" // "@" // tmproot,
		    proc="none", check="absent", index="", extname="", \
		    extver="", ikparams="", omit="", replace="", \
		    outfile=tmpout, logfile="", glogpars="", verbose-)
		if (0 < gemextn.fail_count || nfiles != gemextn.count) {
		    printlog ("ERROR - NSOFFSET: Bad output files.", \
			l_logfile, verbose+) 
		    status = 2
		    goto clean
		}
	    }

	} else {
	    touch (tmpout)
	}


	join (tmproot // "," // tmpout, "", output=tmplist, delim=" ", \
	    missing="Missing", maxchar=1000, shortest-, verbose-)


	first = yes
	scanin1 = tmplist

	while (fscan (scanin1, imgin, imgout) != EOF) {

	    if (debug) print ("testing " // imgin)

	    delete (tmpin, verify-, >& "dev$null")
	    tmpin = mktemp ("tmpin")
	    gemextn (imgin, proc="expand", check="exists", index="", \
		extname=l_sci_ext, extver="1-", ikparams="", omit="", \
		replace="", outfile=tmpin, logfile="dev$null", glogpars="", 
		verbose-)
	    if (0 == gemextn.fail_count && 0 < gemextn.count) {
		scanin2 = tmpin
		junk = fscan (scanin2, imgsci)
		havemdf = no

		if (debug) print ("have science data: " // imgsci)

	    } else {
		delete (tmpin, verify-, >& "dev$null")
		tmpin = mktemp ("tmpin")
		gemextn (imgin, proc="expand", check="table", index="0-", \
		    extname="", extver="", ikparams="", omit="", \
		    replace="%\[%.fits[%", outfile=tmpin, logfile="dev$null",
		    glogpars="", verbose-)
		# ignore errors since many extns not tables
		if (1 == gemextn.count) {
		    scanin2 = tmpin
		    junk = fscan (scanin2, mdfin)
		    havemdf = yes

		    if (debug) print ("have mdf: " // mdfin)

		} else {
		    printlog ("ERROR - NSOFFSET: " // imgin // " missing \
			data/MDF.", l_logfile, verbose+)
		    goto clean
		}
	    }


	    if (l_fl_measure && (first || l_fl_all)) {

		if (debug) print ("need to measure offset")

		if (havemdf) {

		    if (debug) print ("construct image from mdf")

		    if (isindef (scale)) {
			if (debug) print ("finding pixscale")
			hselect (refphu, l_key_pixscale, yes) | scan (scale)
			if (isindef (scale)) scale = l_pixscale
			if (isindef (scale)) {
			    printlog ("ERROR - NSOFFSET: No header for pixel \
				scale (" // l_key_pixscale // ")", \
				l_logfile, verbose+) 
			    printlog ("                  in " // refphu, \
				l_logfile, verbose+) 
			    goto clean
			}
			if (debug) print ("pixscale: " // scale)
		    }

		    if (debug) print ("base image of zeroes")
		    imdelete (tmpmdfimg, verify-, >& "dev$null")
		    tmpmdfimg = mktemp ("tmpmdfimg")
		    imarith (refsci, "*", "0", tmpmdfimg)

		    if (debug) print ("getting table size")
		    tinfo (mdfin, ttout-, >& "dev$null")
		    if (tinfo.tbltype != "fits") {
			printlog ("ERROR - NSOFFSET: Bad MDF in " // mdfin \
			    // ".", l_logfile, verbose+)	
			goto clean
		    } else {
			nsecns = tinfo.nrows
		    }
		    if (debug) print (nsecns // " sections")

		    for (isecn = 1; isecn <= nsecns; isecn += 1) {
			corner = 0
			specorder = INDEF
			tprint (mdfin, prparam-, prdata+, showrow-, \
			    showhdr-, showunits-, \
			    col="x_ccd,y_ccd,slittype,slitid,slitpos_mx, \
				slitpos_my,slitsize_mx,slitsize_my, \
				slittilt_m,slitsize_mr,slitsize_mw,specorder, \
				corner", rows=isecn, pwidth=160) \
			    | scan (x_ccd,y_ccd,slittype,slitid,slitpos_mx, \
				slitpos_my,slitsize_mx,slitsize_my,\
				slittilt_m,slitsize_mr,slitsize_mw,specorder, \
				corner)
			if (isindef (specorder)) {
			    printlog ("ERROR - NSOFFSET: No data found in " \
				// mdfin // ".", l_logfile, verbose+)
			    goto clean
			}
			if (debug) 
			    print ("(" // x_ccd // "," // y_ccd // ") + " \
				// slitpos_mx // "; " // slitsize_mx \
				// "; " // corner)

			# calculation sfrom nscut
			x_ccd = x_ccd + slitpos_mx / scale
			x1 = x_ccd - 0.5 * slitsize_mx / scale
			x2 = x_ccd + 0.5 * slitsize_mx / scale
			# reduce to accomodate
			x1 = x1 - border
			x2 = x2 + border
			y1 = 1
			y2 = ny
			if (x1 < 1) x1 = 1
			if (x2 > nx) x2 = nx
			if (y1 < 1) y1 = 1
			if (y2 > ny) y2 = ny
			if (debug)
			    print ("["//x1//":"//x2//","//y1//":"//y2//"]")

			imdel (tmparith, verify-, >& "dev$null")
			tmparith = mktemp ("tmparith")
			if (0 != corner) {
			    imexpr ("I>="//x1//"+J*"//corner/(y2-y1)\
				//"&&I<="//x2//"-("//y2//"-J)*"//corner\
				    //"/"//(y2-y1)\
				//"&&J>="//y1//"&&J<="//y2//"?1:a", \
				tmparith, tmpmdfimg, 
				dims="auto", intype="auto", outtype="real", \
				refim="auto", bwidth=0, btype="nearest", \
				bpixval=0, rangecheck+, verbose=debug)
			} else {
			    imexpr ("I>="//x1//"&&I<="//x2//"&&J>="//y1\
				//"&&J<="//y2//"?1:a", tmparith, tmpmdfimg, \
				dims="auto", intype="auto", outtype="real", \
				refim="auto", bwidth=0, btype="nearest", \
				bpixval=0, rangecheck+, verbose=debug)
			}
			imdel (tmpmdfimg, verify-, >& "dev$null")
			tmpmdfimg = tmparith
			tmparith = ""
		    }
		    if (debug) print ("mdf generated: " // tmpmdfimg)

		    if (debug) display (tmpmdfimg, 1, zsc-, zr+)
		    if (debug) print ("smoothing mdf image")
		    imdel (tmparith, verify-, >& "dev$null")
		    tmparith = mktemp ("tmparith")
		    gauss (tmpmdfimg, tmparith, blur, ratio=1, theta=0, \
			nsigma=4, bilinear+, boundary="constant", constant=0)
		    imdel (tmpmdfimg, verify-, >& "dev$null")
		    target = tmparith
			
		} else {

		    if (debug) print ("zeroing bad pixels")
		    delete (tmpdq, verify-, >& "dev$null")
		    tmpdq = mktemp ("tmpdq")
		    gemextn (imgin, proc="expand", check="exists", index="", \
			extname=l_dq_ext, extver="1-", ikparams="", omit="", \
			replace="", outfile=tmpdq, logfile="dev$null", 
			glogpars="", verbose-)
		    if (0 != gemextn.fail_count || 0 == gemextn.count) {
			target = imgsci
		    } else {
			scanin2 = tmpdq
			junk = fscan (scanin2, imgdq)
			imdel (tmpblank, verify-, >& "dev$null")
			tmpblank = mktemp ("tmpblank")
			imexpr ("a>0?0:b", tmpblank, imgdq, imgsci, \
			    dims="auto", intype="auto", outtype="auto", \
			    refim="auto", bwidth=0, btype="nearest", \
			    bpixval=0, rangecheck+, verbose=debug)
			target = tmpblank
		    }

		    if (debug) print ("clipping at zero")
		    imdel (tmparith, verify-, >& "dev$null")
		    tmparith = mktemp ("tmparith")
		    imexpr ("max(0,a)", tmparith, target, \
			dims="auto", intype="auto", outtype="auto", \
			refim="auto", bwidth=0, btype="nearest", \
			bpixval=0, rangecheck+, verbose=debug)
		    delete (tmpblank, verify-, >& "dev$null")
		    tmpblank = tmparith; tmparith = ""; target = tmpblank

		}

		if (debug) print ("tapering target")
		imdel (tmptaptag, verify-, >& "dev$null")
		tmptaptag = mktemp ("tmptaptag")
		taper (target, tmptaptag, width="10 %", subtract="none", \
		    function="cosbell", verbose-, >& "dev$null")
		if (debug) display (tmptaptag, 2, zsc-, zr+)

		if (debug) print ("cross-correlating")
		imdel (tmpconv, verify-, >& "dev$null")
		tmpconv = mktemp ("tmpconv")
		printlog ("NSOFFSET: Calculating cross correlation \
		    (please wait).", l_logfile, l_verbose)
		crosscor (tmptapref, tmptaptag, tmpconv, inreal1+, inimag1-, \
		    inreal2+, inimag2-, outreal+, outimag-, coord_shift-, \
		    center+, chop+, pad+, inmemory+, len_blk=256, \
		    verbose=debug)

		if (debug) print ("normalization")
		if (debug) imstat (tmpconv)
		if (debug) imstat (tmpconv, fields="min,max", lower=INDEF, \
		    upper=INDEF, nclip=0, lsigma=3, usigma=3, binwidth=0.1, \
		    format-, cache-) 
		imstat (tmpconv, fields="min,max", lower=INDEF, upper=INDEF, \
		    nclip=0, lsigma=3, usigma=3, binwidth=0.1, format-, \
		    cache-) | scan (cmin, cmax)
		if (cmax > cmin) {
		    imdel (tmparith, verify-, >& "dev$null")
		    tmparith = mktemp ("tmparith")
		    imexpr ("(a-"//cmin//")/"//(cmax-cmin), tmparith, \
			tmpconv, dims="auto", intype="auto", outtype="auto", \
			refim="auto", bwidth=0, btype="nearest", \
			bpixval=0, rangecheck+, verbose=debug)
		    imdelete (tmpconv, verify-, >& "dev$null")
		    tmpconv = tmparith
		    tmparith = ""
		}
		if (debug) display (tmpconv, 3, zsc-, zr-, z1=0, z2=1)
		if (debug) imstat (tmpconv)

		if (debug) print ("filtering")
		imdel (tmparith, verify-, >& "dev$null")
		tmparith = mktemp ("tmparith")
		if (1 == l_axis) {
		    fnx = 1
		    fny = max (min (10, ny), ny / 10)
		} else {
		    fnx = max (min (10, nx), nx / 10)
		    fny = 1
		}
		if (debug) print (l_axis // " " // fnx // " " // fny)
		median (tmpconv, tmparith, fnx, fny, zloreject=INDEF, \
		    zhireject=INDEF, boundary="constant", constant=0, \
		    verbose=debug)
		imdelete (tmpconv, verify-, >& "dev$null")
		tmpconv = tmparith
		tmparith = ""
		if (debug) display (tmpconv, 4, zsc-, zr+)
		if (debug) imstat (tmpconv)

		if (debug) print ("finding peak")
		if (1 == l_axis) defkey = "j"
		else 		 defkey = "k"
		if (l_fl_inter) {
		    printlog ("NSOFFSET: Fit peak with " // defkey \
			// " key.", l_logfile, verbose+)
		    if (1 == l_axis) {
			printlog ("          on line " // (ny/2), \
			    l_logfile, verbose+)
		    } else {
			printlog ("          on column " // (nx/2), \
			    l_logfile, verbose+)
		    }
		    imexam (tmpconv, frame=1, output="dev$null", ncoutput=1, \
			nloutput=1, logfile=tmplog, keeplog+, defkey=defkey, \
			autoredraw+, allframes+, nframes=0, ncstat=5, \
			nlstat=5, graphcur="", imagecur="", \
			wcs="logical", xformat="", yformat="", \
			display="display(image='$1',frame=$2,zsc-)", \
			use_display+, >& "dev$null")
		} else {	    
		    delete (tmpcursor, verify-, >& "dev$null")
		    tmpcursor = mktemp ("tmpcursor")
		    delete (tmplog, verify-, >& "dev$null")
		    tmplog = mktemp ("tmplog")
		    
		    # find rough estimate of peak position
		    minmax (tmpconv, >& "dev$null")
		    printf ("%s\n", minmax.maxpix) | scanf ("[%d,%d]", xc, yc)
		    if (1 == l_axis)
		        yc = int(ny/2)
		    else
		        xc = int(nx/2)
		    
		    # find peak position with Gaussian fit
		    printf (":sigma %d\n", l_sigma, >& tmpcursor)
		    printf (":%6.2f %6.2f %s\n", \
		        real(xc), real(yc), defkey, >>& tmpcursor)
		    imexam (tmpconv, frame=1, output="dev$null", ncoutput=1, \
			nloutput=1, logfile=tmplog, keeplog+, defkey=defkey, \
			autoredraw+, allframes+, nframes=0, ncstat=5, \
			nlstat=5, graphcur="", imagecur=tmpcursor, \
			wcs="logical", xformat="", yformat="", display="", \
			use_display-, >& "dev$null", >G "dev$null")
		}
		l_shift = INDEF
		scanin2 = tmplog
		while (fscanf (scanin2, \
		    "Lines %s center=%f", sjunk, l_shift) != EOF) {
		}
		if (isindef (l_shift)) {
		    printlog ("ERROR - NSOFFSET: Cannot find peak.", \
			l_logfile, verbose+)
		    goto clean
		}
		if (debug) print ("centre: " // l_shift)

		if (debug) print ("calculate offset")
		# see help crosscor for explanation
		hselect (tmpconv, "CRPIX" // l_axis, yes) | scan (crpix)
		if (1 == l_axis) l_shift = l_shift - crpix
		else		 l_shift = l_shift - crpix

		if (debug) print ("centre: " // l_shift)
		# restrict to 0.1 pix resoln
		l_shift = 0.1 * nint (10 * l_shift)
		printlog ("NSOFFSET: Offset for " // imgin // " is " \
		    // l_shift // ".", l_logfile, l_verbose)
	    }


	    if (l_fl_apply) {

		if (debug) print ("copying input to output")
		if (debug) print (imgin // " -> " // imgout)
		copy (imgin // ".fits", imgout // ".fits", verbose=debug)

		if (havemdf) {

		    if (debug) print ("finding table in output")
		    delete (tmptab, verify-, >& "dev$null")
		    tmptab = mktemp ("tmptab")
		    gemextn (imgout, proc="expand", check="table", \
			index="0-", extname="", extver="", ikparams="", \
			omit="", replace="%\[%.fits[%", outfile=tmptab, \
			logfile="dev$null", glogpars="", verbose-)
		    scanin2 = tmptab
		    junk = fscan (scanin2, mdfout)

		    if (debug) print ("updating table")
		    if (1 == l_axis) {
			tcalc (mdfout, outcol="x_ccd", \
			    equals="x_ccd-" // l_shift, datatype="real")
			printlog ("NSOFFSET: Shifting " // mdfout \
			    // " in X by " // l_shift // " pixels.", \
			    l_logfile, l_verbose) 
		    } else {
			tcalc (mdfout, outcol="y_ccd", \
			    equals="y_ccd-" // l_shift, datatype="real")
			printlog ("NSOFFSET: Shifting " // mdfout \
			    // " in Y by " // l_shift // " pixels.", \
			    l_logfile, l_verbose)
		    }

		} else {

		    if (debug) print ("copying reference wcs")
		    gemwcscopy (imgout // "[0]", refphu, verbose-, \
			logfile=l_logfile)

		    if (debug) print ("shifting wcs")
		    hselect (imgout // "[0]", "CRPIX" // l_axis, yes) \
			| scan (crpix)
		    hedit (imgout // "[0]", "CRPIX" // l_axis, \
			(crpix+l_shift), add+, addonly-, delete-, verify-, \
			show=debug, update+)
		    if (1 == l_axis) {
			printlog ("NSOFFSET: Shifting " // imgout \
			    // " WCS in X by " // l_shift // " pixels.", \
			    l_logfile, l_verbose) 
		    } else {
			printlog ("NSOFFSET: Shifting " // imgout \
			    // " WCS in Y by " // l_shift // " pixels.", \
			    l_logfile, l_verbose)
		    }

		}
	    }

	    first = no
	}

	status = 0

clean:

	shift = l_shift

	if (0 != status && 2 != status && access (tmpout)) {
	    delete ("@" // tmpout, verify-, >& "dev$null")
	    imdelete ("@" // tmpout, verify-, >& "dev$null")
	}

	delete (tmpref, verify-, >& "dev$null")
	delete (tmprefsci, verify-, >& "dev$null")
	delete (tmproot, verify-, >& "dev$null")
	delete (tmpout, verify-, >& "dev$null")
	delete (tmplist, verify-, >& "dev$null")
	delete (tmpin, verify-, >& "dev$null")
	delete (tmpdq, verify-, >& "dev$null")
	imdelete (tmpmdfimg, verify-, >& "dev$null")
	imdelete (tmparith, verify-, >& "dev$null")
	imdelete (tmptapref, verify-, >& "dev$null")
	imdelete (tmptaptag, verify-, >& "dev$null")
	imdelete (tmpconv, verify-, >& "dev$null")
	imdelete (tmpblank, verify-, >& "dev$null")
	imdelete (tmpmask, verify-, >& "dev$null")
	delete (tmpcursor, verify-, >& "dev$null")
	delete (tmplog, verify-, >& "dev$null")
	delete (tmptab, verify-, >& "dev$null")

	scanin1 = ""
	scanin2 = ""

	printlog (" ", l_logfile, l_verbose) 
	if (0 == status) {
	    printlog ("NSOFFSET exit status:  good.", l_logfile, l_verbose) 
	} else {
	    printlog ("NSOFFSET exit status:  error.", l_logfile, l_verbose) 
	}
	printlog ("---------------------------------------------------------\
	    ----------------------", l_logfile, l_verbose) 
	
end
