# Copyright(c) 2001-2004 Association of Universities for Research in Astronomy, Inc.

procedure nsslitfunction (skyflats, dark, flat, output) 

# makes an illumination corrected GNIRS/NIRI spectroscopic flat field
# 
# Version    Date            Author      Comment 
#   1.0      2000/03/24      bmiller     created from nsflat
#            Apr 6, 2000     IJ          new wmef
#   1.1      2000/04/07      bmiller     addmask for gemdqadd, error checking
#   1.2      2000/04/25                  logfile, verbosity, less disk use
#   1.3      2000/05/08                  fix specsec checking
#            2000/06/26                  check output not ""
#            2003/08/19      KL          IRAF2.12 - new/modified parameters
#                                          hedit: addonly
#                                          imcombine: headers,bpmasks,expmasks
#                                              outlimits, rejmask->rejmasks,
#                                              plfile->nrejmasks
#            2003/10/29      KL          moved from niri to gnirs package

string	skyflats	{"", prompt = "Input sky flats"}
string	dark		{"", prompt = "Input dark image"}
string	flat		{"", prompt = "Input lamp flat"} 
string	output		{"", prompt = "Output normalized flat (MEF)"}
string	title		{"", prompt = "Title for output SCI plane"}
string	combine		{"median", prompt = "Combination operation"}
string	reject		{"avsigclip", prompt = "Rejection algorithm"}
string	scale		{"none", prompt = "Image scaling"}
string	zero		{"mean", prompt = "Image zeropoint offset"}
string	weight		{"none", prompt = "Image weights"}
string	statsec		{"[*,*]", prompt = "Statistics section"}
real	lthreshold	{INDEF, prompt = "Lower threshold"}
real	hthreshold	{INDEF, prompt = "Upper threshold"}
int	nlow		{1, min=1, prompt = "minmax: Number of low pixels to reject"}
int	nhigh		{1, min=1, prompt = "minmax: Number of high pixels to reject"}
int	nkeep		{0, min=0, prompt = "Minimum to keep or maximum to reject"}  
bool	mclip		{yes, prompt = "Use median in sigma clipping algorithms?"}
real	lsigma		{3., prompt = "Lower sigma clipping factor"}
real	hsigma		{3., prompt = "Upper sigma clipping factor"}
real	ron		{0.0, min = 0., prompt = "Readout noise rms in electrons"}
real	gain		{1.0, min = 0.00001, prompt = "Gain in e-/ADU"}
string	snoise		{"0.0", prompt = "ccdclip: Sensitivity noise (electrons)"}
real	sigscale	{0.1, prompt = "Tolerance for sigma clipping scaling correction"}
real	pclip		{-0.5, prompt = "pclip: Percentile clipping parameter"}
real	grow		{0.0, prompt = "Radius (pixels) for neighbor rejection"}
string	function	{"spline3", prompt = "Fitting function"}
int	order		{3, min = 1, prompt = "Order of fitting function"}
bool	fl_vary		{no, prompt = "Allow scaling to vary with order?"}
bool	fl_dqprop	{no, prompt = "Propagate all DQ values?"}
bool	fl_inter	{no, prompt = "Fit interactively?"}
string	logfile		{"", prompt = "Logfile name"}
bool	verbose		{yes, prompt = "Verbose output?"}
int	status		{0, prompt = "Exit status (0=good)"}
struct	*scanfile	{"", prompt = "Internal use only"} 

begin

	string	l_skyflats = ""
	string	l_dark = ""
	string	l_flat = ""
	string	l_output = ""
	string	l_title = ""
	string	l_combine = ""
	string	l_reject = ""
	string	l_scale = ""
	string	l_zero = ""
	string	l_weight = ""
	string	l_statsec = ""
	real	l_lthreshold
	real	l_hthreshold
	int	l_nlow
	int	l_nhigh
	int	l_nkeep
	bool	l_mclip
	real	l_lsigma
	real	l_hsigma
	real	l_ron
	real	l_gain
	string	l_snoise = ""
	real	l_sigscale
	real	l_pclip
	real	l_grow
	string	l_function = ""
	int	l_order
	bool	l_fl_vary
	bool	l_fl_dqprop
	bool	l_fl_inter
	string	l_logfile = ""
	bool	l_verbose

	string	l_sci_ext
	string	l_var_ext
	string	l_dq_ext
	string	l_key_ron
	string	l_key_gain
	string	l_key_dispaxis
	string	l_key_exptime

        string	filelist, scilist, dqlist, specsec1, sflat, phulist
        string	img, dqsum, dqsumold, suf, response, illumvar, tmplog, phu
        string	tmpdq, combout, combsig, combdq, flatsub, illum, badhdr
        real	norm, stddev, firstexp
        int	i, k, n, j, l, idum, nbad, len, junk, nsci
	string	keyfound
        struct	sdate, line
	bool	debug, first
	string	ssrclist, sdestlist, vsrclist, vdestlist, comblist, scratch
	string	ssrc, sdest, vsrc, vdest
	real	mdresp, mddata, ratio, correcn, oldcorrecn, oldmddata

        
	junk = fscan (	skyflats, l_skyflats)
	junk = fscan (	dark, l_dark)
	junk = fscan (	flat, l_flat)
	junk = fscan (	output, l_output)
	junk = fscan (	title, l_title)
	junk = fscan (	combine, l_combine)
	junk = fscan (	reject, l_reject)
	junk = fscan (	scale, l_scale)
	junk = fscan (	zero, l_zero)
	junk = fscan (	weight, l_weight)
	junk = fscan (	statsec, l_statsec)
	l_lthreshold =	lthreshold
	l_hthreshold =	hthreshold
	l_nlow =	nlow
	l_nhigh =	nhigh
	l_nkeep =	nkeep
	l_mclip =	mclip
	l_lsigma =	lsigma
	l_hsigma =	hsigma
	l_ron =		ron
	l_gain =	gain
	junk = fscan (	snoise, l_snoise)
	l_sigscale =	sigscale
	l_pclip =	pclip
	l_grow =	grow
	junk = fscan (	function, l_function)
	l_order =	order
	l_fl_vary =	fl_vary
	l_fl_dqprop =	fl_dqprop
	l_fl_inter =	fl_inter
	junk = fscan (	logfile, l_logfile)
	l_verbose =	verbose
        
	badhdr = ""
	junk = fscan (	nsheaders.sci_ext, l_sci_ext)
	if ("" == l_sci_ext) badhdr = badhdr + " sci_ext"
	junk = fscan (	nsheaders.var_ext, l_var_ext)
	if ("" == l_var_ext) badhdr = badhdr + " var_ext"
	junk = fscan (	nsheaders.dq_ext, l_dq_ext)
	if ("" == l_dq_ext) badhdr = badhdr + " dq_ext"
	junk = fscan (	nsheaders.key_ron, l_key_ron)
	if ("" == l_key_ron) badhdr = badhdr + " key_ron"
	junk = fscan (	nsheaders.key_gain, l_key_gain)
	if ("" == l_key_gain) badhdr = badhdr + " key_gain"
	junk = fscan (	nsheaders.key_dispaxis, l_key_dispaxis)
	if ("" == l_key_dispaxis) badhdr = badhdr + " key_dispaxis"
	junk = fscan (	nsheaders.key_exptime, l_key_exptime)
	if ("" == l_key_exptime) badhdr = badhdr + " key_exptime"

	status = 1
	debug = no
        
        cache ("keypar", "hedit", "gnirs", "gemextn")

        filelist = mktemp ("tmpfilelist") 
        phulist = mktemp ("tmpphulist") 
	if (debug) print (phulist)
        scilist = mktemp ("tmpscilist") 
        combout = mktemp ("tmpcombout") 
        flatsub = mktemp ("tmpflatsub") 
        sflat = mktemp ("tmpsflat") 
        response = mktemp ("tmpresponse") 
	ssrclist = mktemp ("tmpssrclist")
	sdestlist = mktemp ("tmpsdestlist")
	vsrclist = mktemp ("tmpvsrclist")
	vdestlist = mktemp ("tmpvdestlist")
	comblist = mktemp ("tmpcomblist")
	scratch = mktemp ("tmpscratch")
 

	# Start logging

	if ("" == l_logfile) {
	    junk = fscan (gnirs.logfile, l_logfile) 
	    if (l_logfile == "") {
		l_logfile = "gnirs.log"
		printlog ("WARNING - NSSLITFUNCTION: Both nsslitfunc.logfile \
		    and gnirs.logfile are empty.", l_logfile, verbose+) 
		printlog ("                   Using default file " \
		    // l_logfile // ".", l_logfile, verbose+) 
	    }
	}
	date | scan (sdate) 
	printlog ("-----------------------------------------------------\
	    -------------------------", l_logfile, verbose = l_verbose) 
	printlog ("NSSLITFUNC -- " // sdate, l_logfile, l_verbose) 
	printlog (" ", l_logfile, l_verbose) 
        printlog ("skyflats = " // l_skyflats, l_logfile, l_verbose) 
        printlog ("dark = " // l_dark, l_logfile, l_verbose) 
        printlog ("output = " // l_output, l_logfile, l_verbose) 
        printlog ("title = " // l_title, l_logfile, l_verbose) 
        printlog ("function = " // l_function, l_logfile, l_verbose) 
        printlog ("order = " // l_order, l_logfile, l_verbose) 
        printlog ("sci_ext = " // l_sci_ext, l_logfile, l_verbose) 
        printlog ("var_ext = " // l_var_ext, l_logfile, l_verbose) 
        printlog ("dq_ext = " // l_dq_ext, l_logfile, l_verbose) 
        printlog (" ", l_logfile, l_verbose) 
        
	if ("" != badhdr) {
            printlog ("ERROR - NSSLITFUNCTION: Parameter(s) missing from \
		nsheaders: " // badhdr, l_logfile, verbose+) 
            goto clean
	}


	gemextn (l_skyflats, check = "exists,mef", process = "none", \
	    index = "", extname = "", extversion = "", ikparams = "", \
	    omit = "", replace = "", outfile = filelist, logfile="",
	    glogpars="", verbose=l_verbose)
	if (gemextn.fail_count != 0 || gemextn.count == 0) {
	    printlog ("ERROR - NSSLITFUNCTION: Bad input images.", l_logfile, 
		verbose+) 
	    goto clean
	}
	n = gemextn.count
	gemextn (l_skyflats, check = "exists,mef", process = "expand", \
	    index = "0", extname = "", extversion = "", ikparams = "", \
	    omit = "", replace = "", outfile = phulist, logfile="",
	    glogpars="", verbose=l_verbose)
	if (gemextn.fail_count != 0 || gemextn.count != n) {
	    printlog ("ERROR - NSSLITFUNCTION: Bad or missing PHUs.", \
		l_logfile, verbose+) 
	    goto clean
	}
	head (filelist, nlines=1) | scan (img)
	gemextn (img, check = "", process = "expand", index = "", \
	    extname = l_sci_ext, extversion = "1-", ikparams = "", \
	    omit = "", replace = "", outfile = "dev$null", \
	    logfile="dev$null", glogpars="", verbose-)
	nsci = gemextn.count


	gemextn (l_dark, check = "exists,mef", process = "none", \
	    index = "", extname = "", extversion = "", ikparams = "", \
	    omit = "", replace = "", outfile = "dev$null", logfile="",
	    glogpars="", verbose=l_verbose)
	if (gemextn.fail_count != 0 || gemextn.count != 1) {
	    printlog ("ERROR - NSSLITFUNC: Bad or missing dark.", l_logfile, 
		verbose+) 
	    goto clean
	}
	gemextn (l_dark, check = "exists,mef", process = "none", \
	    index = "", extname = "", extversion = "", ikparams = "", \
	    omit = "", replace = "", outfile = "STDOUT", \
	    logfile="dev$null", glogpars="", verbose-) | scan (line)
	l_dark = line
	gemextn (l_dark, check = "", process = "expand", index = "", \
	    extname = l_sci_ext, extversion = "1-", ikparams = "", \
	    omit = "", replace = "", outfile = "dev$null", \
	    logfile="dev$null", glogpars="", verbose-)
	if (nsci != gemextn.count) {
	    printlog ("ERROR - NSSLITFUNCTION: Skyflats have different no \
		of extensions (" // nsci // ")", l_logfile, verbose+) 
	    printlog ("                        to combined dark (" \
		// gemextn.count // ")", l_logfile, verbose+) 
	    goto clean
	}
        

	gemextn (l_flat, check = "exists,mef", process = "none", \
	    index = "", extname = "", extversion = "", ikparams = "", \
	    omit = "", replace = "", outfile = "dev$null", logfile="",
	    glogpars="", verbose=l_verbose)
	if (gemextn.fail_count != 0 || gemextn.count != 1) {
	    printlog ("ERROR - NSSLITFUNC: Bad or missing flat.", l_logfile, 
		verbose+) 
	    goto clean
	}
	gemextn (l_flat, check = "exists,mef", process = "none", \
	    index = "", extname = "", extversion = "", ikparams = "", \
	    omit = "", replace = "", outfile = "STDOUT", \
	    logfile="dev$null", glogpars="", verbose-) | scan (line)
	l_flat = line
	gemextn (l_flat, check = "", process = "expand", index = "", \
	    extname = l_sci_ext, extversion = "1-", ikparams = "", \
	    omit = "", replace = "", outfile = "dev$null", \
	    logfile="dev$null", glogpars="", verbose-)
	if (nsci != gemextn.count) {
	    printlog ("ERROR - NSSLITFUNCTION: Skyflats have different no \
		of extensions (" // nsci // ")", l_logfile, verbose+) 
	    printlog ("                        to input flat (" \
		// gemextn.count // ")", l_logfile, verbose+) 
	    goto clean
	}


	gemextn (l_output, check = "absent", process = "none", \
	    index = "", extname = "", extversion = "", ikparams = "", \
	    omit = "", replace = "", outfile = "dev$null", logfile="",
	    glogpars="", verbose=l_verbose)
	if (gemextn.fail_count != 0 || gemextn.count != 1) {
	    printlog ("ERROR - NSSLITFUNC: Bad or existing output.", \
		l_logfile, verbose+) 
	    goto clean
	}
	gemextn (l_output, check = "absent", process = "none", \
	    index = "", extname = "", extversion = "", ikparams = "", \
	    omit = "", replace = "", outfile = "STDOUT", \
	    logfile="dev$null", glogpars="", verbose-) | scan (line)
	l_output = line


	firstexp = INDEF
	scanfile = phulist
	while (fscan (scanfile, phu) != EOF) {

	    keyfound = ""
	    hselect(phu, "*PREPAR*", yes) | scan (keyfound)
	    if (keyfound == "") {
		printlog ("ERROR - NSSLITFUNCTION: Image " // phu \
		    // " not PREPAREd.", l_logfile, verbose+) 
		goto clean
	    }
	    
	    keypar (phu, l_key_exptime, silent+)
	    if (keypar.found) {
		if (isindef (firstexp)) {
		    firstexp = real (keypar.value)
		} else {
		    if (abs (firstexp - real (keypar.value)) > 0.1) {
			printlog ("ERROR - NSSLITFUNCTION: Exposure times \
			    vary: " // firstexp // "/" // keypar.value \
			    // ".", l_logfile, verbose+) 
		    }
		}
	    } else {
		printlog ("ERROR - NSSLITFUNCTION: Image " // phu \
		    // " has no " // l_key_exptime, l_logfile, verbose+) 
		goto clean
	    }

	}


	# Combine sky flats

	printlog ("NSSLITFUNCTION: Combining sky flats.", \
	    l_logfile, verbose+)
	gemcombine ("@" // filelist, combout, title="", combine=l_combine, \
	    reject=l_reject, offsets="none", scale=l_scale, zero=l_zero, \
	    weight=l_weight, statsec=l_statsec, expname=l_key_exptime, \
	    lthreshold=l_lthreshold, hthreshold=l_hthreshold, nlow=l_nlow, \
	    nhigh=l_nhigh, nkeep=l_nkeep, mclip=l_mclip, lsigma=l_lsigma, \
	    hsigma=l_hsigma, key_ron=l_key_ron, key_gain=l_key_gain, \
	    ron=l_ron, gain=l_gain, snoise=l_snoise, sigscale=l_sigscale, \
	    pclip=l_pclip, grow=l_grow, bpmfile="", nrejfile="", \
	    sci_ext=l_sci_ext, var_ext=l_var_ext, dq_ext=l_dq_ext, \
	    fl_vardq+, logfile=l_logfile, fl_dqprop=l_fl_dqprop, \
	    verbose-)

	if (gemcombine.status != 0) {
            printlog ("ERROR - NSSLITFUNCTION: Error combining sky flats.", \
		l_logfile, verbose+)
            goto clean
	}


	# Subtract dark

	if (debug) print ("subtracting dark")
	gemarith (combout, "-", l_dark, flatsub, sci_ext=l_sci_ext, \
	    var_ext=l_var_ext, dq_ext=l_dq_ext, fl_vardq+, \
	    logfile=l_logfile, verbose-)
	if (! imaccess (flatsub)) {
            printlog ("ERROR - NSSLITFUNCTION: Error subtracting dark.", \
		l_logfile, verbose+)
            goto clean
	}


	# Divide by lamp flat

	if (debug) print ("dividing by lamp flat")
	gemarith (flatsub, "/", l_flat, sflat, sci_ext=l_sci_ext, \
	    var_ext=l_var_ext, dq_ext=l_dq_ext, fl_vardq+, \
	    logfile=l_logfile, verbose-)
	if (! imaccess (sflat)) {
            printlog ("ERROR - NSSLITFUNCTION: Error dividing by lamp flat.", \
		l_logfile, verbose+)
            goto clean
	}
        

	# Do the illumination thing to each version

	gemextn (sflat, check = "exists,mef", process = "expand", \
	    index = "", extname = l_sci_ext, extversion = "1-", \
	    ikparams = "", omit = "", replace = "", outfile = scilist, \
	    logfile="", glogpars="", verbose=l_verbose)
	if (gemextn.fail_count != 0 || gemextn.count == 0) {
	    printlog ("ERROR - NSSLITFUNCTION: Missing internal data.", \
		l_logfile, verbose+) 
	    goto clean
	}
	n = gemextn.count

	copy (l_flat // ".fits", l_output // ".fits")
	gemextn (l_output, check = "", process = "expand", index = "", \
	    extname = l_sci_ext, extversion = "1-", ikparams = "overwrite", \
	    omit = "", replace = "", outfile = ssrclist, logfile="",
	    glogpars="", verbose=l_verbose)
	if (gemextn.fail_count != 0 || gemextn.count != n) {
	    printlog ("ERROR - NSSLITFUNCTION: Sky flats inconsistent with \
		lamp flats (cut before processing).", l_logfile, verbose+) 
	    goto clean
	}
	gemextn (l_output, check = "", process = "expand", index = "", \
	    extname = l_sci_ext, extversion = "1-", ikparams = "overwrite", \
	    omit = "", replace = "", outfile = sdestlist, logfile="",
	    glogpars="", verbose=l_verbose)
	gemextn (l_output, check = "", process = "expand", index = "", \
	    extname = l_var_ext, extversion = "1-", ikparams = "overwrite", \
	    omit = "", replace = "", outfile = vsrclist, logfile="",
	    glogpars="", verbose=l_verbose)
	gemextn (l_output, check = "", process = "expand", index = "", \
	    extname = l_var_ext, extversion = "1-", ikparams = "overwrite", \
	    omit = "", replace = "", outfile = vdestlist, logfile="",
	    glogpars="", verbose=l_verbose)

	joinlines (scilist // "," // ssrclist // "," // sdestlist // "," \
	    // vsrclist // "," // vdestlist, "", output=comblist, delim=" ", \
	    missing = " ", maxchars=1000, shortest-, verbose-)

	scanfile = comblist
	vsrc = ""
	vdest = ""
	first = yes

	while (fscan (scanfile, img, ssrc, sdest, vsrc, vdest) != EOF) {

	    if (debug)
		imstat (img, fields="image,npix,mean,stddev,min,max,midpt")
	    if (debug) print ("illumination")
	    illumination (img, response, low_rej=2., high_rej=2, \
                niter=3, order=l_order, nbins=5, function=l_function, \
                interpolator="poly3", interactive=l_fl_inter)
	    if (debug)
		imstat (response, \
		    fields="image,npix,mean,stddev,min,max,midpt")

	    imstat (img, fields="midpt", lower=INDEF, upper=INDEF, \
		nclip=0, lsigma=3, usigma=3, binwidth=0.1, format-,
		cache-) | scan (mddata)
	    imstat (response, fields="midpt", lower=INDEF, upper=INDEF, \
		nclip=0, lsigma=3, usigma=3, binwidth=0.1, format-,
		cache-) | scan (mdresp)

	    # if fl_vary, scaling is just 1/response to get unity
	    # otherwise, that's only true for the first extension
	    # and later extensions must reflect the change in mddata
	    if (fl_vary) {
		correcn = 1.0 / mdresp
	    } else {
		if (first) {
		    oldmddata = mddata
		    correcn = 1.0 / mdresp
		} else {
		    correcn = (1.0 / mdresp) * (mddata / oldmddata)
		}
	    }
	    if (debug) print ("correction: " // correcn)
	    imarith (response, "*", correcn, response, verbose=debug)

	    if (debug) print ("scaling sci")
	    if (debug) imhead (ssrc)
	    if (debug) imhead (response)
	    imarith (ssrc, "*",  response, scratch, verbose-)
	    imcopy (scratch, sdest, >& "dev$null")
	    imdel (scratch, ver-, >& "dev$null")
	    scratch = mktemp ("tmpscratch")
	    if (imaccess (vsrc)) {
		if (debug) print ("scaling var")
		imarith (vsrc, "*",  response, scratch, verbose-)
		imarith (scratch, "*",  response, scratch, verbose-)
		imcopy (scratch, vdest, >& "dev$null")
		imdel (scratch, ver-, >& "dev$null")
		scratch = mktemp ("tmpscratch")
		imdel (response, ver-, >& "dev$null")
		response = mktemp ("tmpresponse")
	    }

	    first = no
	}
        

        date | scan (sdate) 

	if (debug) print (l_output)
        hedit (l_output // "[0]", "NSSLIT", sdate, add+, addonly-, delete-, \
	    verify-, show-, update+) 
        if (l_title != "") {
	    hedit (l_output // "[0]", "i_title", l_title, add+, addonly-, \
		delete-, verify-, show-, update+) 
        }

	status = 0


clean:

	if (debug) print ("clean")
	delete (filelist, verify-, >& "dev$null")
	delete (phulist, verify-, >& "dev$null")
	delete (scilist, verify-, >& "dev$null")
	delete (ssrclist, verify-, >& "dev$null")
	delete (sdestlist, verify-, >& "dev$null")
	delete (vsrclist, verify-, >& "dev$null")
	delete (vdestlist, verify-, >& "dev$null")
	delete (comblist, verify-, >& "dev$null")

	imdelete (combout, verify-, >& "dev$null")
	imdelete (flatsub, verify-, >& "dev$null")
	imdelete (sflat, verify-, >& "dev$null")
	imdelete (response, verify-, >& "dev$null")
	imdelete (scratch, verify-, >& "dev$null")

        printlog (" ", l_logfile, l_verbose) 
        printlog ("NSSLITFUNCTION done", l_logfile, l_verbose) 
        printlog ("------------------------------------------------------\
	    --------------------------", l_logfile, l_verbose)
        
end
