# Cpyright(c) 2001-2004 Association of Universities for Research in Astronomy, Inc.

procedure nstransform (inimages) 

# Rectify GNIRS/NIRI spectra using available wavelength calibrations
# and S-distoritions.
#
# Version: Sept 20, 2002 IJ,MT,JJ v1.4 release
#          Aug 19, 2003  KL  IRAF2.12 - new parameter, addonly, in hedit
#          Oct 29, 2003  KL  moved from niri to gnirs package

char	inimages	{prompt = "Input GNIRS/NIRI spectra"}						# OLDP-1-primary-single-prefix=t
char	outspectra	{"", prompt = "Output spectra"}							# OLDP-1-output
char	outprefix	{"t", prompt = "Prefix for output spectra"}					# OLDP-4
char	lamptransf	{"", prompt = "Name of arc lamp spectrum used to determine wavelength calibration"}			# OLDP-1-input
char	sdisttransf	{"", prompt = "Name of image used to determine S-distortion calibration"}	# OLDP-1-input
int	dispaxis	{1, min = 1, max = 2, prompt = "Dispersion axis if not defined in the header"}	# OLDP-3
char	database	{"", prompt = "Directory for calibration files"}				# OLDP-3
bool	fl_inter	{no, prompt = "Examine fitted transformations interactively"}			# OLDP-4
bool	fl_stripe	{no, prompt = "Stripe in spatial direction (for XD wavelength reference)?"}	# OLDP-3
bool	fl_align	{no, prompt = "Align images in the spatial direction?"}				# OLDP-3
char	function	{"chebyshev", min = "legendre|chebyshev|spline1|spline3", prompt = "Coordinate fitting function"}	# OLDP-2
int	xorder		{2, min = 1, prompt = "X order of coordinate fitting function"}			# OLDP-2
int	yorder		{2, min = 1, prompt = "Y order of coordinate fitting function"}			# OLDP-2
char	interptype	{"poly3", enum = "nearest|linear|poly3|poly5|spline3", prompt = "Interpolation type for transform"}	# OLDP-2
bool	xlog		{no, prompt = "Logarithmic x coordinate for transform"}				# OLDP-2
bool	ylog		{no, prompt = "Logarithmic y coordinate for transform"}				# OLDP-2
real	pixscale	{1., prompt = "Pixel scale in arcsec/pixel"}					# OLDP-3
char	logfile		{"", prompt = "Logfile"}							# OLDP-1
bool	verbose		{yes, prompt = "Verbose"}							# OLDP-2
bool	force		{no, prompt = "Force use with earlier IRAF versions?"}				# OLDP-3
int	status		{0, prompt = "Exit status (0=Good)"}						# OLDP-4
struct*	scanfile1	{"", prompt = "Internal use"}							# OLDP-4
struct*	scanfile2	{"", prompt = "Internal use"}							# OLDP-4


begin
        
        char	l_inimages = ""
	char	l_outspectra = ""
	char	l_outprefix = ""
	char	l_lamptransf = ""
	char	l_sdisttransf = ""
	int	l_dispaxis
	char	l_database = ""
	bool	l_fl_inter
	bool	l_fl_stripe
	bool	l_fl_align
	char	l_function = ""
	int	l_xorder
	int	l_yorder
	char	l_interptype = ""
	bool	l_xlog
	bool	l_ylog
	real	l_pixscale
	char	l_logfile = ""
	bool	l_verbose
	bool	l_force

        char	l_key_dispaxis = ""
        char	l_sci_ext = ""
	char	l_var_ext = ""
	char	l_dq_ext = ""
	char	l_key_xoff = ""
	char	l_key_yoff = ""
	char	l_key_pixscale = ""

	char	badhdr, name
        char      tmpinlist, tmpoutlist, tmpin, tmpphu, tmpsci
        char      tmplamps, tmpsdist, tmpdqint, tmpdqint2, s_database
        int       nlamps, nsdist, ii, nver, version
        bool      l_fl_lamptransf, l_fl_sdisttransf
        char      l_fitname, l_fitname_lamp, l_fitname_sdist
        struct    l_date
        char      wmefimages, l_extnames, lampsec, sdistsec
	char	lampsec2, sdistsec2, line
        char      tmpfile1, img, imgin, imgout, lamp, sdist, sec, phu, key
	char	sec2, sec4, refimg
        int       nbad, nfiles, lcount, junk, iextn, nx, ny
        bool      fl_run_before, debug, first
	char	extn[3], tmpmedian
	real	dx, dy, x1, x2, y1, y2, pixsc, shift, shiftx, shifty, small

	char	tmpfit, dbfitfile, keyfound
	bool	beforesurface
	struct 	sline
	int	count, nfail, naxis
	real	coeff
        
        # Define temporary files
        tmpfile1 = mktemp ("tmpfile1") 
        tmpin = mktemp ("tmpin") 
        tmpphu = mktemp ("tmpphu") 
        tmpsci = mktemp ("tmpsci") 
        tmpinlist = mktemp ("tmpinl") 
        tmpoutlist = mktemp ("tmpout") 
        tmplamps = mktemp ("tmplam") 
        tmpsdist = mktemp ("tmpsdi") 
        
        # Initialize exit status
        status = 1
        debug = no

	small = 0.001
        
        # cache parameter files
        cache ("niri", "hedit", "gemextn", "gemoffsetlist", "gemwcscopy") 
        
        # Set the local variables

	junk = fscan (	inimages, l_inimages)
	junk = fscan (	outspectra, l_outspectra)
	junk = fscan (	outprefix, l_outprefix)
	junk = fscan (	lamptransf, l_lamptransf)
	junk = fscan (	sdisttransf, l_sdisttransf)
	junk = fscan (	database, l_database)
	junk = fscan (	function, l_function)
	junk = fscan (	interptype, l_interptype)
	junk = fscan (	logfile, l_logfile)

	l_dispaxis =	dispaxis
	l_fl_inter =	fl_inter
	l_fl_stripe =	fl_stripe
	l_fl_align =	fl_align
	l_xorder =	xorder
	l_yorder =	yorder
	l_xlog =	xlog
	l_ylog =	ylog
	l_pixscale =	pixscale
	l_verbose =	verbose
	l_force =	force

	badhdr = ""
	junk = fscan (	nsheaders.key_dispaxis, l_key_dispaxis)
	if ("" == l_key_dispaxis) badhdr = badhdr + " key_dispaxis"
	junk = fscan (	nsheaders.sci_ext, l_sci_ext)
	if ("" == l_sci_ext) badhdr = badhdr + " sci_ext"
	junk = fscan (	nsheaders.var_ext, l_var_ext)
	if ("" == l_var_ext) badhdr = badhdr + " var_ext"
	junk = fscan (	nsheaders.dq_ext, l_dq_ext)
	if ("" == l_dq_ext) badhdr = badhdr + " dq_ext"
        junk = fscan (	nsheaders.key_xoff, l_key_xoff)
	if ("" == l_key_xoff) badhdr = badhdr + " key_xoff"
        junk = fscan (	nsheaders.key_yoff, l_key_yoff)
	if ("" == l_key_yoff) badhdr = badhdr + " key_yoff"
        junk = fscan (	nsheaders.key_pixscale, l_key_pixscale)
	if ("" == l_key_pixscale) badhdr = badhdr + " key_pixscale"

	extn[1] = l_sci_ext
	extn[2] = l_var_ext
	extn[3] = l_dq_ext

        if (l_logfile == "") {
            junk = fscan (gnirs.logfile, l_logfile)
            if (l_logfile == "") {
                l_logfile = "gnirs.log"
                printlog ("WARNING - NSTRANSFORM:  Both nstransform.logfile \
		    and gnirs.logfile", l_logfile, verbose+) 
                printlog ("                        are empty.  \
		    Using " // l_logfile, l_logfile, verbose+) 
            }
        }


        printlog ("-------------------------------------------------------------------------------", l_logfile, l_verbose) 
        date | scan (l_date) 
        printlog ("NSTRANSFORM -- " // l_date, l_logfile, l_verbose) 
        printlog ("", l_logfile, l_verbose) 

        
        if (l_database == "") {
            junk = fscan (gnirs.database, l_database)
            if (l_database == "") {
                l_database = "database"
                printlog ("WARNING - NSTRANSFORM:  Both nstransform.database \
		    and gnirs.database", l_logfile, verbose+) 
                printlog ("                        are empty.  \
		    Using " // l_database, l_logfile, verbose+) 
            }
        }


        s_database = ""
        if (l_database != "") {
	    s_database = l_database
	    if (strldx ("/", s_database) != strlen (s_database)) {
		s_database = s_database // "/"
	    }
	}

        
	if ("" != badhdr) {
            printlog ("ERROR - NSTRANSFORM: Parameter(s) missing from \
		nsheaders: " // badhdr, l_logfile, verbose+) 
            goto clean
	}
 

	# Expand and verify input
        
	if (debug) print ("expansion and verification of input")

	gemextn (l_inimages, check="mef,exists", process="none", index="", \
	    extname="", extversion="", ikparams="", omit="extension, kernel", \
	    replace="", outfile=tmpfile1, logfile="", glogpars="",
	    verbose=l_verbose)
	if (0 != gemextn.fail_count || 0 == gemextn.count) {
	    printlog ("ERROR - NSTRANSFORM:  Problems with input.", \
		l_logfile, verbose+)
	    goto clean
	}
	nfiles = gemextn.count


	# Check PHUs

	gemextn ("@" // tmpfile1, check="mef,exists", process="append", \
	    index="0", extname="", extversion="", ikparams="", \
	    omit="extension, kernel", replace="", outfile=tmpphu, 
	    logfile="", glogpars="", verbose=l_verbose)
	if (nfiles != gemextn.count || 0 != gemextn.fail_count) {
	    printlog ("ERROR - NSTRANSFORM:  Problems with PHUs.", l_logfile, \
		verbose+)
	    goto clean
	}

        scanfile1 = tmpphu
        scanfile2 = tmpfile1
	nbad = 0
	nfiles = 0
        while (fscan (scanfile1, phu) != EOF) {

	    junk = fscan (scanfile2, img)
	    if (debug) print ("file: " // phu)
	    fl_run_before = no

	    hselect (phu, "NSTRANSF", yes) | scan (line)
	    if (1 == nscan ()) {
		printlog ("WARNING - NSTRANSFORM: Image " // img // " has \
		    been run", l_logfile, verbose+) 
		printl2og ("                       through NSTRANSFORM \
		    before.", l_logfile, verbose+) 
		fl_run_before = yes
	    }

	    keyfound=""
	    hselect(phu, "*PREPAR*", yes) | scan(keyfound)
	    if (keyfound == "") {
		printlog ("WARNING - NSTRANSFORM: Image " // img // " has \
		    not been PREPAREd", l_logfile, verbose+) 
		nbad = nbad + 1
	    }

	    if (no == fl_run_before) {
		print (img, >> tmpinlist)
		nfiles = nfiles + 1
	    }

	}
        
        
	# Check SCIs

	gemextn ("@" // tmpfile1, check="mef,exists", process="expand", \
	    index="", extname=l_sci_ext, extversion="1-", ikparams="", \
	    omit="extension,index", replace="", outfile=tmpsci, \
	    logfile="", glogpars="", verbose=l_verbose)
	if (0 == gemextn.count || 0 != gemextn.fail_count) {
	    printlog ("ERROR - NSTRANSFORM:  Problems with science data.", \
		l_logfile, verbose+)
	    goto clean
	}

        scanfile1 = tmpsci
        while (fscan (scanfile1, img) != EOF) {
	    hselect (img, "NAXIS", yes) | scan (naxis)
	    if (1 != nscan()) {
		printlog ("WARNING - NSTRANSFORM: Data in " // img \
		    // " missing NAXIS.", l_logfile, verbose+) 
		nbad = nbad + 1
	    } else if (2 != naxis) {
		printlog ("WARNING - NSTRANSFORM: Data in " // img \
		    // " are not 2-D.", l_logfile, verbose+) 
		nbad = nbad + 1
	    }
	}
        
        
        # Check for empty file list

        if (nfiles == 0) {
            printlog ("ERROR - NSTRANSFORM:  No input images to process.", 
                l_logfile, verbose+) 
            goto clean
        }
        

        # Exit if problems found with input files

        if (nbad > 0) {
            printlog ("ERROR - NSTRANSFORM: " // nbad // " bad image(s)", \
		l_logfile, verbose+) 
            goto clean
        }
        
        printlog ("Using input files:", l_logfile, l_verbose) 
        if (l_verbose) type (tmpinlist) 
        type (tmpinlist, >> l_logfile) 
        delete (tmpfile1, ver-) 
        

	# Expand and verify output
	
 	if (debug) print ("expansion and verification of output")

	gemextn (l_outspectra, check="absent", process="none", index="", \
	    extname="", extversion="", ikparams="", omit="extension, kernel", \
	    replace="", outfile=tmpoutlist, logfile="", glogpars="",
	    verbose=l_verbose)
	if (0 != gemextn.fail_count) {
	    printlog ("ERROR - NSTRANSFORM:  Problems with output.", \
		l_logfile, verbose+)
	    goto clean
	}
	if (0 == gemextn.count) {
	    gemextn ("%^%" // l_outprefix // "%" // "@" // tmpinlist, \
		check="absent", process="none", index="", \
		extname="", extversion="", ikparams="", omit="kernel,exten", \
		replace="", outfile=tmpoutlist, logfile="", glogpars="",
		verbose=l_verbose)
	    if (gemextn.fail_count != 0 || gemextn.count == 0) {
		printlog ("ERROR - NSTRANSFORM: No or incorrectly formatted \
		    output files", l_logfile, verbose+) 
		goto clean
	    }
        }

        printlog ("Using output files:", l_logfile, l_verbose) 
        if (l_verbose) type (tmpoutlist) 
        type (tmpoutlist, >> l_logfile) 
        

        # Check number of input and output images

        if (nfiles != gemextn.count) {
            printlog ("ERROR - NSTRANSFORM: Different number of input and \
		output files", l_logfile, verbose+) 
            goto clean
        }
        
        
	# Check that all images have the same number of SCI
	# extensions.

	# Compare against the first input image

	if (debug) print ("checking version numbers")

	head (tmpinlist, nlines=1) | scan (refimg)
	gemextn (refimg, check="exists", process="expand", index="", \
	    extname=l_sci_ext, extversion="1-", ikparams="", omit="", \
	    replace="", outfile="dev$null", logfile="", glogpars="",
	    verbose-)
	if (0 != gemextn.fail_count || 0 == gemextn.count) {
	    printlog ("ERROR - NSTRANSFORM: Bad science data in " // refimg \
		// ".", l_logfile, verbose+)
	    goto clean
	}
	# Assume they're 1..n - this is checked by the extver range below
	nver = int (gemextn.count)

	scanfile1 = tmpinlist
	while (fscan (scanfile1, img) != EOF) {
	    gemextn (img, check="exists", process="expand", index="", \
		extname=l_sci_ext, extversion="1-" // nver, \
		ikparams="", omit="", replace="", outfile="dev$null", \
		logfile="", glogpars="", verbose-)
	    if (0 != gemextn.fail_count || nver != gemextn.count) {
		printlog ("ERROR - NSTRANSFORM: Bad or missing science data \
		    in " // img // ".", l_logfile, verbose+)
		goto clean
	    }
	}

	
        # Check number of lamps and sdist calibration files.
	# Check if they exist.
        
        # First check the wavelength calibration files

	l_fl_lamptransf = no
        if (l_lamptransf != "") {

	    if (debug) print ("checking lamp file")

	    gemextn (l_lamptransf, check="", process="none", index="", \
		extname="", extversion="", ikparams="", \
		omit="extension, kernel", replace="", outfile=tmplamps, \
		logfile="", glogpars="", verbose=l_verbose)
	    if (0 != gemextn.fail_count) {
		printlog ("ERROR - NSTRANSFORM:  Problems with lamp files.", \
		    l_logfile, verbose+)
		goto clean
	    }

	    nlamps = gemextn.count
            l_fl_lamptransf = (nlamps > 0)

	    if (no == l_fl_lamptransf) {
		printlog ("WARNING - NSTRANSFORM:  lamptransf non-null, but \
		    no lamp files")
	    }

	}

	if (l_fl_lamptransf) {

	    if (debug) print ("checking number of lamp files")

            if (nlamps != 1 && nlamps != nfiles) {     
                printlog ("ERROR - NSTRANSFORM: Number of wavelength \
		    calibration files is " // nlamps // ". It must be \
		    either 1", l_logfile, verbose+) 
                printlog ("                     or equal to the number of \
		    input spectra (" // nfiles // ").", l_logfile, verbose+) 
                goto clean
            }
            
            scanfile1 = tmplamps
	    nbad = 0
            while (fscan (scanfile1, img) != EOF) {
		for (version = 1; version <= nver; version = version + 1) {
		    sec = "_" // l_sci_ext // "_" // version // "_"
		    name = s_database // "id" // img // sec
		    if (no ==access (name) ) {
			printlog ("WARNING - NSTRANSFORM: wavelength \
			    calibration file ", l_logfile, verbose+)
			printlog ("                       " // name \
			    // " does not exist.", l_logfile, verbose+) 
			nbad = nbad + 1
		    }
		}
            }
	    if (nbad == nver) {
		printlog ("ERROR - NSTRANSFORM: no wavelength information.", \
		    l_logfile, verbose+)
		goto clean
	    }
            
        }

        
        # Second check the distortion correction files

	l_fl_sdisttransf = no
        if (l_sdisttransf != "") {

	    if (debug) print ("checking s-dist file")

	    gemextn (l_sdisttransf, check="", process="none", index="", \
		extname="", extversion="", ikparams="", \
		omit="extension, kernel", replace="", outfile=tmpsdist, \
		logfile="", glogpars="", verbose=l_verbose)
	    if (0 != gemextn.fail_count) {
		printlog ("ERROR - NSTRANSFORM:  Problems with sdist files.", \
		    l_logfile, verbose+)
		goto clean
	    }

	    nsdist = gemextn.count
            l_fl_sdisttransf = (nsdist > 0)

	    if (no == l_fl_sdisttransf) {
		printlog ("WARNING - NSTRANSFORM:  sdisttransf non-null, but \
		    no sdist files")
	    }

	}

	if (l_fl_sdisttransf) {

	    if (debug) print ("checking number of s-dist files")

            if (nsdist != 1 && nsdist != nfiles) {     
                printlog ("ERROR - NSTRANSFORM: Number of wavelength \
		    calibration files is " // nsdist // ". It must be \
		    either 1", l_logfile, verbose+) 
                printlog ("                     or equal to the number of \
		    input spectra (" // nfiles // ").", l_logfile, verbose+) 
                goto clean
            }
            
            scanfile1 = tmpsdist
	    nbad = 0
            while (fscan (scanfile1, img) != EOF) {
		for (version = 1; version <= nver; version = version + 1) {
		    sec = "_" // l_sci_ext // "_" // version // "_"
		    name = s_database // "id" // img // sec
		    if (no ==access (name) ) {
			printlog ("WARNING - NSTRANSFORM: sdist \
			    calibration file ", l_logfile, verbose+) 
			printlog ("                       " // name \
			    // " does not exist.", l_logfile, verbose+) 
			nbad = nbad + 1
		    }
		}
            }
	    if (nbad == nver) {
		printlog ("ERROR - NSTRANSFORM: no sdistortion information.", \
		    l_logfile, verbose+)
		goto clean
	    }
            
        } else {

	    if (l_fl_align) {
		printlog ("WARNING - NSTRANSFORM: no spatial transform, so \
		    setting fl_align=no", l_logfile, verbose+) 
		l_fl_align = no
	    }

	}
       

        if (no == l_fl_lamptransf && no == l_fl_sdisttransf) {
            printlog ("ERROR - NSTRANSFORM: No wavelength or S-distortion \
                transformations defined", l_logfile, verbose+) 
            goto clean
        }
        

	if (l_fl_align) {
	    gemoffsetlist ("@" // tmpinlist, refimg, distance = 0.0, \
		age = INDEF, fl_younger+, fl_nearer+, fl_noref-, \
		targetlist = "dev$null", offsetlist = "dev$null", \
		logfile = l_logfile, verbose = debug, direction = 3, \
		key_xoff = l_key_xoff, key_yoff = l_key_yoff, key_time = "",
		force = l_force)

	    if (gemoffsetlist.status != 0) {
		printlog ("ERROR - NSTRANSFORM: Error in GEMOFFSETLIST.  \
		    Can't determine offsets.", l_logfile, verbose+) 
		goto clean
	    }
	}



        # At this point tmplamps contains the wavelength calibration files and
        #               tmpsdist contains the s-distortion calibration files.
        

        # Add input, output lamps and sdist to the input list, matched with the
        # correct images

 	if (debug) print ("constructing main file list")

        join (tmpinlist, tmpoutlist, delim = " ", missing = "Missing", 
            maxchar = 161, shortest = yes, verbose = no, > tmpin) 
        delete (tmpinlist // "," // tmpoutlist, ver-, >& "dev$null") 
        
        if (l_fl_lamptransf) {
            if (nlamps == 1) {
                scanfile1 = tmpin
		head (tmplamps, nlines=1) | scan (l_lamptransf)
                while (fscan (scanfile1, sline) != EOF) {
                    print (sline // " " // l_lamptransf, >> tmpinlist) 
                }
            } else {
                join (tmpin, tmplamps, output = tmpinlist, delim = " ", \
		    missing = "Missing", maxchar = 161, shortest+, verbose-) 
            } 
        } else {
            scanfile1 = tmpin
            while (fscan (scanfile1, sline) != EOF) {
                print (sline // " none ", >> tmpinlist) 
            }
        }
        delete (tmpin, ver-, >& "dev$null") 
        rename (tmpinlist, tmpin, field = "all") 
        
        if (l_fl_sdisttransf) {
            if (nsdist == 1) {
                scanfile1 = tmpin
		head (tmpsdist, nlines=1) | scan (l_sdisttransf)
                while (fscan (scanfile1, sline) != EOF) {
                    print (sline // " " // l_sdisttransf, >> tmpinlist) 
                }
            } else {
                join (tmpin, tmpsdist, output = tmpinlist, delim = " ", \
		    missing = "Missing", maxchar = 161, shortest+, verbose-) 
            } 
        } else {
            scanfile1 = tmpin
            while (fscan (scanfile1, sline) != EOF) {
                print (sline // " none ", >> tmpinlist) 
            }
        }
        delete (tmpin, ver-, >& "dev$null") 
        rename (tmpinlist, tmpin, field = "all") 
        delete (tmplamps // "," // tmpsdist, verify-, >>& "dev$null") 
        


        # Main loop for transformation
        
        lcount = 0
        scanfile1 = tmpin
        while (fscan (scanfile1, imgin, imgout, lamp, sdist) != EOF) {
            
	    if (debug) print ("processing " // imgin)
	    phu = imgin // "[0]"
	    l_dispaxis = dispaxis
	    hselect (phu, l_key_dispaxis, yes) | scan (l_dispaxis)

	    nfail = 0

	    pixsc = l_pixscale
	    hselect (phu, l_key_pixscale, yes) | scan (pixsc)

	    if (small > pixsc && l_fl_align) {

		printlog ("WARNING - NSTRANSFORM: Zero pixscale, no \
		    alignment for " // imgin, l_logfile, verbose+) 

	    } else if (l_fl_align) {

		hselect (phu, l_key_xoff, yes) | scan (shiftx)
		if (0 == nscan ()) {
		    printlog ("WARNING - NSTRANSFORM: No X offset for " \
			// phu, l_logfile, verbose+) 
		    shiftx = 0.0
		}

		hselect (phu, l_key_yoff, yes) | scan (shifty)
		if (0 == nscan ()) {
		    printlog ("WARNING - NSTRANSFORM: No X offset for " \
			// phu, l_logfile, verbose+) 
		    shifty = 0.0
		}

		if ((1 == l_dispaxis && small < abs (shiftx)) ||
		    (2 == l_dispaxis && small < abs (shifty))) {
		    printlog ("WARNING - NSTRANSFORM: Non-zero shift in \
			dispersion direction for", l_logfile, verbose+) 
		    printlog ("                       " // imgin \
			// " (ignoring)", l_logfile, verbose+) 
		}

		if (1 == l_dispaxis) shift = shifty / pixsc
		else shift = shiftx / pixsc
		if (debug) print ("shifty " // shifty)
		if (debug) print ("shiftx " // shiftx)
		if (debug) print ("shift " // shift // ", " // pixsc)

		printlog ("NSTRANSFORM: Aligning " // imgin // " by " \
		    // shift // " pixels", l_logfile, l_verbose)

	    }

	    for (version = 1; version <= nver; version = version + 1) {

		if (debug) print ("version " // version)

		if (1 == version) first = yes
		sec = "_" // l_sci_ext // "_" // version // "_"
		sec2 = "[" // l_sci_ext // "," // version // "]"
		sec4 = "[" // l_sci_ext // "," // version // ",inherit]"

		lampsec = lamp // sec
		sdistsec = sdist // sec
		lampsec2 = lamp // sec2
		sdistsec2 = sdist // sec2

		if (debug) print ("checking for files")

		if ((l_fl_lamptransf && \
		     (! access (s_database // "id" // lampsec))) || \
		    (l_fl_sdisttransf && \
		     (! access (s_database // "id" // sdistsec)))) {

		    printlog ("WARNING - NSTRANSFORM: Incomplete \
			transform information for " // imgin // sec, \
			l_logfile, verbose+) 
		    nfail = nfail + 1

		} else {

		    if (debug) print ("using files")

		    if (debug) print (imgin // sec4)
		    if (debug) print ("dispaxis: " // l_dispaxis) 

		    longslit.dispaxis = l_dispaxis
		    dx = INDEF; dy = INDEF


		    # Fit wavelength solution to input image

		    l_fitname_lamp = ""
		    if (l_fl_lamptransf) {

			l_fitname_lamp = imgout // "lamp"
			if (debug) print ("fitting wavelength to " \
			    // lampsec2 // ", " // l_fitname_lamp) 

			delete (s_database // "fc" // l_fitname_lamp, \
			    verify-, >& "dev$null")

			# juggle file names around here
			del (s_database // "id" // lamp, \
			    verify-, >& "dev$null")
			copy (s_database // "id" // lampsec, \
			    s_database // "id" // lamp)
			if (debug) print ("fitcoords...")
			fitcoords (lampsec2, fitname = l_fitname_lamp, \
			    interactive = l_fl_inter, combine = yes, \
			    database = l_database, deletions = "", \
			    function = l_function, xorder = l_xorder, \
			    yorder = l_yorder, logfiles = l_logfile, \
			    plotfile = "")
			delete (s_database // "id" // lamp, \
			    verify-, >& "dev$null")
			if (debug) print ("...done")

		    }


		    # Fit the s-distortion solution to input image

		    l_fitname_sdist = ""
		    if (l_fl_sdisttransf) {

			l_fitname_sdist = imgout // "sdist"
			if (debug) print ("fitting spatial to " \
			    // sdistsec // ", " // l_fitname_sdist) 

			delete (s_database // "fc" // l_fitname_sdist, \
			    verify-, >& "dev$null")
			
			# juggle file names around here
			del (s_database // "id" // sdist, \
			    verify-, >& "dev$null")
			copy (s_database // "id" // sdistsec, \
			    s_database // "id" // sdist)
			if (debug) print ("fitcoords...")
			fitcoords (sdistsec2, fitname = l_fitname_sdist, \
			    interactive = l_fl_inter, combine = yes, \
			    database = l_database, deletions = "", \
			    function = l_function, xorder = l_xorder, \
			    yorder = l_yorder, logfiles = l_logfile, \
			    plotfile = "") 
			delete (s_database // "id" // sdist, \
			    verify-, >& "dev$null")
			if (debug) print ("...done")


			if (small < pixscale && l_fl_align) {

			    tmpfit = mktemp ("tmpfit")
			    dbfitfile = s_database // "fc" // imgout // "sdist"

			    rename (dbfitfile, tmpfit)

			    scanfile2 = tmpfit
			    beforesurface = yes
			    count = 0
			    while (fscan (scanfile2, sline) != EOF) {
				if (beforesurface) {
				    beforesurface = \
					strstr ("surface", sline) == 0
				} else {
				    count = count + 1
				    if (count == 9) {
					coeff = real (sline) + shift
					printf ("%f\n", coeff) | scan (sline)
				    }
				}
				print (sline, >> dbfitfile)
			    }

			    delete (tmpfit, verify-, >& "dev$null")
			}
		    }

		    # If order along spatial direction = 1 then we
		    # need to fix the scale (only correcting shape, 
		    # not scale)
		    # But only on final fit!  If we do this for spatial
		    # alone then the wavelength cal cannot be combined
		    # because the x size is different.
		    if (l_fl_lamptransf && l_fl_sdisttransf) {
			if (1 == l_dispaxis) {
			    if (1 == l_yorder) {
				dy = 1
				printlog ("WARNING - NSTRANSFORM: Keeping \
				    spatial (y) scale constant for " \
				    // imgin // sec, l_logfile, l_verbose) 
			    }
			} else if (2 == l_dispaxis) {
			    if (1 == l_xorder) {
				dx = 1
				printlog ("WARNING - NSTRANSFORM: Keeping \
				    spatial (x) scale constant for " \
				    // imgin // sec, l_logfile, l_verbose) 
			    }
			}
		    }

		    l_fitname = l_fitname_lamp // "," // l_fitname_sdist
		    printlog ("NSTRANSFORM: Fitnames for " // imgin \
			// ": " // l_fitname, l_logfile, l_verbose) 


		    if (first) {
			if (debug) print ("copying phu")
			imcopy (imgin // "[0]", imgout, verbose-) 
			first = no
		    }


		    # Image size in spatial direction

		    img = imgin // "[" // l_sci_ext // "," // version // "]"
		    if (debug) print ("i_naxis" // (3 - l_dispaxis))
		    hselect (img, "i_naxis" // (3 - l_dispaxis), yes) \
			| scan (junk)
		    if (1 == l_dispaxis) {
			nx = 1
			ny = junk
			x1 = INDEF
			x2 = INDEF
			y1 = 1
			y2 = ny
		    } else {
			nx = junk
			ny = 1
			x1 = 1
			x2 = nx
			y1 = INDEF
			y2 = INDEF
		    }
		    

		    # Transform the data using the fit information

		    # Copy to local header - don't want to use "inherit"
		    # because we're messing with WCS and the PHU has 
		    # its own WCS that might confuse things.

		    if (debug) print (l_key_dispaxis // " " // l_dispaxis)
		    img = imgin // "[" // l_sci_ext // "," // version // "]"
		    hedit (img, l_key_dispaxis, l_dispaxis, add+, addonly-, \
			delete-, verify-, show-, update+)

		    if (debug) print ("transform sci") 

		    # Reduce this to avoid mis-matches and errors
		    # By this point in the process all images will be 2D,
		    # although some may have started larger
		    hedit (img, "WCSDIM", 2, add+, \
			addonly-, delete+, verify-, show-, update+) 

		    transform (img, imgout // "[" // l_sci_ext // \
			"," // version // ",append]", fitnames = l_fitname, \
			database = l_database, interptype = l_interptype, \
			xlog = l_xlog, ylog = l_ylog, flux = yes, \
			logfiles = l_logfile, dx=dx, dy=dy, \
			x1=x1, x2=x2, y1=y1, y2=y2) 

		    if (fl_stripe) {
		        tmpmedian = mktemp ("tmpmedian") 
			imcopy (imgout // "[" // l_sci_ext // \
			    "," // version // "]", tmpmedian, >& "dev$null")
			if (debug) print ("striping " // nx // "," // ny)
			median (tmpmedian, tmpmedian, \
			    nx, ny, boundary = "wrap", zlo = 0.001, \
			    zhi = INDEF, constant = 0, >& "dev$null")
			imcopy (tmpmedian, imgout // "[" \
			    // l_sci_ext // "," // version // ",overwrite]", \
			    >& "dev$null")
			imdelete (tmpmedian, ver-, >& "dev$null")
		    }

		    img = imgin // "[" // l_var_ext // "," // version // "]"
		    if (imaccess (img)) {
			if (debug) print ("transform var") 
			hedit (img, l_key_dispaxis, l_dispaxis, add+, \
			    addonly-, delete-, verify-, show-, update+)
			hedit (img, "WCSDIM", 2, add+, \
			    addonly-, delete+, verify-, show-, update+) 
			transform (img, imgout // "[" // l_var_ext // "," \
			    // version // ",append]", fitnames = l_fitname, \
			    database = l_database, interptype = l_interptype, \
			    xlog = l_xlog, ylog = l_ylog, flux = yes, \
			    logfiles = l_logfile, dx=dx, dy=dy, \
			    x1=x1, x2=x2, y1=y1, y2=y2) 
		    }

		    img = imgin // "[" // l_dq_ext // "," // version // "]"
		    if (imaccess (img)) {
			if (debug) print ("transform dq") 
			tmpdqint = mktemp ("tmpdqintimg") 

			# transform (must be linear to avoid bleeding)
			hedit (img, l_key_dispaxis, l_dispaxis, add+, \
			    addonly-, delete-, verify-, show-, update+)
			hedit (img, "WCSDIM", 2, add+, \
			    addonly-, delete+, verify-, show-, update+) 
			transform (img, tmpdqint, fitnames = l_fitname, \
			    database = l_database, interptype = "linear", \
			    xlog = l_xlog, ylog = l_ylog, flux = yes, \
			    logfiles = l_logfile, dx=dx, dy=dy, \
			    x1=x1, x2=x2, y1=y1, y2=y2) 

			# anything non-zero has been contaminated
			# so set to 1 (all we can do, really)
			imexpr ("a > 0.00001 ? 1 : 0", \
			    imgout // "[" // l_dq_ext // "," \
			    // version // ",append]", tmpdqint, \
			    dims = "auto", intype = "auto", outtype = "int", \
			    refim = "auto", bwidth = 0, btype = "nearest", \
			    bpixval = 0, rangecheck+, verbose-, \
			    exprdb = "none")

			imdelete (tmpdqint, verify-) 
		    }


		    # Copy across nsappwave info if no wavelength transform

		    if (no == l_fl_lamptransf) {
			for (iextn = 1; iextn <= 3; iextn = iextn+1) {

			    sec = "[" // extn[iext] // "," // version // "]"

			    if (imaccess (imgin // sec) && \
				imaccess (imgout // sec)) {

				if (debug) print ("nsappwave info for " // sec)
				if (debug) print ("from " // imgin)
				if (debug) print ("to " // imgout)

				key = "CTYPE" // l_dispaxis
				hselect (imgin // sec, key, yes) \
				    | scan (sline)
				if (fscan (sline, line) > 0) {
				    hedit (imgout // sec, key, \
					value=sline, add+, addonly-, \
					delete-, verify-, show-, update+)
				}

				key = "CRPIX" // l_dispaxis
				hselect (imgin // sec, key, yes) \
				    | scan (sline)
				if (fscan (sline, line) > 0) {
				    hedit (imgout // sec, key, \
					value=sline, add+, addonly-, \
					delete-, verify-, show-, update+)
				}

				key = "CRVAL" // l_dispaxis
				hselect (imgin // sec, key, yes) \
				    | scan (sline)
				if (fscan (sline, line) > 0) {
				    hedit (imgout // sec, key, \
					value=sline, add+, addonly-, \
					delete-, verify-, show-, update+)
				}

				key = "CD" // l_dispaxis // "_" // l_dispaxis
				hselect (imgin // sec, key, yes) \
				    | scan (sline)
				if (fscan (sline, line) > 0) {
				    hedit (imgout // sec, key, \
					value=sline, add+, addonly-, \
					delete-, verify-, show-, update+)
				}

				key = "WAT" // l_dispaxis // "_000"
				hselect (imgin // sec, key, yes) \
				    | scan (sline)
				if (fscan (sline, line) > 0) {
				    hedit (imgout // sec, key, \
					value=sline, add+, addonly-, \
					delete-, verify-, show-, update+)
				}

				key = "WAT" // l_dispaxis // "_001"
				hselect (imgin // sec, key, yes) \
				    | scan (sline)
				if (fscan (sline, line) > 0) {
				    hedit (imgout // sec, key, \
					value=sline, add+, addonly-, \
					delete-, verify-, show-, update+)
				}

				key = "WAT" // l_dispaxis // "_002"
				hselect (imgin // sec, key, yes) \
				    | scan (sline)
				if (fscan (sline, line) > 0) {
				    hedit (imgout // sec, key, \
					value=sline, add+, addonly-, \
					delete-, verify-, show-, update+)
				}
			    }
			}
		    }
		}
	    }

	    # Update the header

	    date | scan (sline) 
	    flpr
	    gemhedit (imgout // "[0]", "NSTRANSF", sline, \
		"Time stamp for NSTRANSFORM") 
	    hedit (imgout // "[0]", "GEM-TLM", sline, add+, addonly-, \
		del-, ver-, show-, update+) 

	    # If aligned, move reference WCS to shifted PHU

	    if (l_fl_align) {
		if (debug)
		    print ("wcs from " // refimg // " to " // imgout)
		# This used to require a patch to work correctly...
		gemwcscopy (imgout // "[0]", refimg // "[0]", verbose-, \
		    logfile=l_logfile)
		# ...so check with gemoffsetlist
		gemoffsetlist (infiles=imgout, reffile=refimg, distance=1, \
		    age=INDEF, targetlist="dev$null", offsetlist="dev$null", \
		    fl_nearer+, direction=3, fl_younger+, fl_noref+, \
		    wcs_source="phu", key_xoff=l_key_xoff, \
		    key_yoff=l_key_yoff, key_time="", logfile=l_logfile, \
		    verbose-, force=l_force)
		if (gemoffsetlist.status != 0 || gemoffsetlist.count != 1) {
		     printlog ("WARNING - NSTRANSFORM: The WCS information \
			in the aligned file has not been", \
			l_logfile, verbose+)
		     printlog ("                       corrected.  Stack the \
			results with gemcombine, not ", \
			l_logfile, verbose+)
		     printlog ("                       nscombine or nsstack \
			(see `help nstransform`).", \
			l_logfile, verbose+)
		}
	    }

	    if (nfail > 0) {
		hselect (imgout // "[0]", "NSCIEXT", yes) | scan (junk)
		hedit (imgout // "[0]", "NSCIEXT", \
		    (junk - nfail), add+, addonly-, delete-, \
		    verify-, show-, update+)
	    }

        }
        
        
	# Completed successfully
	status = 0


clean:
        if (status == 0) {
            printlog ("", l_logfile, no) 
            printlog ("NSTRANSFORM  Exit status good", l_logfile, l_verbose) 
            printlog ("-------------------------------------------------------------\
                ------------------", l_logfile, l_verbose) 
        }
        scanfile1 = ""
        scanfile2 = ""
        delete (tmpin, ver-, >& "dev$null") 
        delete (tmpinlist, ver-, >& "dev$null") 
        delete (tmpoutlist, ver-, >& "dev$null") 
        delete (tmpfile1, ver-, >& "dev$null") 
        delete (tmpphu, ver-, >& "dev$null") 
        delete (tmpsci, ver-, >& "dev$null") 
        delete (tmplamps // "," // tmpsdist, verify-, >>& "dev$null") 
        
end
