# Copyright(c) 2001-2004 Association of Universities for Research in Astronomy, Inc.

procedure mprepare(inimages)

# Procedure to take raw Michelle data, with n MEF extensions (where n is the
# number of nodsets), and reformat the data so that it is consistent with
# the higher level data reduction image format.

# Data format is flipped from difference frame (ext 1), nod A (ext 2), nod
# B (ext 3) to nod A (ext 1), nod B (ext2), difference frame (ext 3).

# Version Sep  3, 2003  TB, first write
#         Oct 29, 2003  KL  IRAF 2.12 - new parameters
#                             hedit: addonly

char  inimages     {prompt="Input Michelle image(s)"}               # OLDP-1-input-primary-single-prefix=n
char  rawpath      {"",prompt="Path for input raw images"}      # OLDP-4
char  outimages    {"",prompt="Output image(s)"}                # OLDP-1-output
char  outprefix    {"m",prompt="Prefix for output image(s)"}    # OLDP-4
char  logfile      {"",prompt="Logfile"}                        # OLDP-1
bool  verbose      {yes,prompt="Verbose"}            # OLDP-4
int   status       {0,prompt="Exit status (0=good)"} # OLDP-4
struct *scanfile   {"",prompt="Internal use only"}   # OLDP-4

begin
    char   l_inimages, l_outimages
    char   l_rawpath, l_prefix, l_logfile, l_temp
    char   tmpim1, tmpim2, tmpfile2, tmpfile, tmphead, tmpim3, tmpim4
    char   in[1000], inpath[1000], out[1000]
    char   header, paramstr
    int    i, nimages, noutimages, maxfiles, nbad, n_j, modeflag
    int    next_inp
    bool   l_verbose,  alreadyfixed[1000], bad
    int    next_sci
    struct l_struct

    status=0
    maxfiles=1000
    # cache imgets - used throughout the script
    cache("imgets")

    # set the local variables
    l_inimages=inimages ; l_outimages=outimages ; l_rawpath=rawpath
    l_verbose=verbose ; l_prefix=outprefix ; l_logfile=logfile

    # open temp files
    tmpfile = mktemp("tmpfile")
    tmpfile2 = mktemp("tmpfile2")
    tmphead = mktemp("tmphead")

    # Create the list of parameter/value pairs.  One pair per line.
    # All lines combined into one string.  Line delimiter is '\n'.
    paramstr =  "inimages       = "//inimages.p_value//"\n"
    paramstr += "rawpath        = "//rawpath.p_value//"\n"
    paramstr += "outimages      = "//outimages.p_value//"\n"
    paramstr += "outprefix      = "//outprefix.p_value//"\n"
    paramstr += "logfile        = "//logfile.p_value//"\n"
    paramstr += "verbose        = "//verbose.p_value
    
    # Assign a logfile name if not given. Open logfile and start log.
    # Write parameter/value pairs ("paramstr") to log.
    gloginit( l_logfile, "mprepare", "midir", paramstr, fl_append+,
        verbose=l_verbose )
    if (gloginit.status != 0) {
        status = gloginit.status
	goto exitnow
    }
    l_logfile = gloginit.logfile.p_value

    #----------------------------------------------------------------------
    # Check the rawpath name for a final /
    if (substr(l_rawpath,(strlen(l_rawpath)),(strlen(l_rawpath))) != "/")
	l_rawpath=l_rawpath//"/"
    if (l_rawpath=="/" || l_rawpath==" ")
	l_rawpath=""

    #-----------------------------------------------------------------------
    # Load up arrays of input name lists
    # This version handles both *s, / and commas in l_inimages

    # check that list file exists
    if (substr(l_inimages,1,1)=="@") {
	l_temp=substr(l_inimages,2,strlen(l_inimages))
	if (!access(l_temp) && !access(l_rawpath//l_temp)) {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=101,
	        str="Input file "//l_temp//" not found.",verbose+)
	    status=1
	    goto clean
	}
    }


    # parse wildcard and comma-separated lists
    if (substr(l_inimages,1,1)=="@")
	scanfile=substr(l_inimages,2,strlen(l_inimages))
    else {
	print(l_inimages, >tmpfile)
	sed("-e",'s/\,/\ /g',tmpfile) | words("STDIN", >tmpfile2)
	scanfile=tmpfile2
	delete(tmpfile,verify-, >& "dev$null")
    }  
    while (fscan(scanfile,l_temp) != EOF) {
	files(l_rawpath//l_temp,sort-) | unique("STDIN", >> tmpfile)
    }

    delete(tmpfile2,verify-, >& "dev$null")
    scanfile=tmpfile

    nimages=0
    nbad=0
    while (fscan(scanfile,l_temp) != EOF) {
	bad=no
	# remove rawpath
	l_temp=substr(l_temp,strlen(l_rawpath)+1,strlen(l_temp))

	# remove .fits if present
	if (substr(l_temp,strlen(l_temp)-4,strlen(l_temp)) == ".fits")
	    l_temp=substr(l_temp,1,(strlen(l_temp)-5))

	# check to see if the file exists (or if it's not MEF FITS)
	gimverify(l_rawpath//l_temp)
	if (gimverify.status==1) {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=101,
	        str="Input image "//l_rawpath//l_temp//" not found.",
		verbose+)
	    nbad+=1
	    bad=yes
	} else if (gimverify.status>1) {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=123,
	        str="Input image not multiple-extension FITS.",verbose+)
	    nbad+=1
	    bad=yes
	} else {
	    nimages=nimages+1
	    if (nimages > maxfiles) {
	        glogprint( l_logfile, "mprepare", "status", type="error",
		    errno=121, str="Maximum number of input images exceeded \
		    ("//str(maxfiles)//")",verbose+)
		status=1
		goto clean
	    }
	    
	    # check to see if already fixed with MPREPARE
	    if (!bad) {
		imgets(l_rawpath//l_temp//"[0]","MPREPARE", >& "dev$null")
		if (imgets.value != "0") {
		    alreadyfixed[nimages]=yes
		    glogprint( l_logfile, "mprepare", "status", type="warning",
		        errno=0, str="Image "//l_rawpath//l_temp//\
		        " already fixed using MPREPARE.",verbose=l_verbose)
		    glogprint( l_logfile, "mprepare", "status", type="string",
		        str="              Data scaling, correction for number \
			of reads, and header",verbose=l_verbose)
		    glogprint( l_logfile, "mprepare", "status", type="string",
		        str="              updating not performed.",
			verbose=l_verbose)
		} else
		    alreadyfixed[nimages]=no
		in[nimages]=l_temp
	    } # end if(!bad)
	    # trim off path from in[i], if present
	    i=strlen(in[nimages])-1
	    while (i>1) {
		if (substr(in[nimages],i,i) == "/") {
		    inpath[nimages]=substr(in[nimages],1,i)
		    in[nimages]=substr(in[nimages],i+1,strlen(in[nimages]))
		    i=0
		}
		i=i-1
	    } # end while(i>1) path-chopping loop
	} # end else
    } # end while(fscan) loop

    # Exit if problems found
    if (nbad > 0) {
        glogprint( l_logfile, "mprepare", "status", type="error", errno=101,
	    str=nbad//" image(s) either do not exist or are not MEF files.",
	    verbose+)
	status=1
	goto clean
    }

    glogprint( l_logfile, "mprepare", "status", type="string",
        str="Processing "//nimages//" files",verbose=l_verbose)
    scanfile=""
    delete(tmpfile,ver-, >& "dev$null")

    # Now for the output images
    # outimages could contain legal * if it is of a form like %st%stX%*.imh

    noutimages=0
    nbad=0
    print(l_outimages) | scan(l_outimages)

    if (l_outimages!="" && l_outimages!=" ") {
	if (substr(l_outimages,1,1)=="@")
	    scanfile=substr(l_outimages,2,strlen(l_outimages))
	else { 
	    files(l_outimages,sort-) | unique("STDIN", > tmpfile)
	    scanfile=tmpfile
	}

	while (fscan(scanfile,l_temp) != EOF) {
	    # remove .fits if present
	    if (substr(l_temp,strlen(l_temp)-4,strlen(l_temp)) == ".fits")
		l_temp=substr(l_temp,1,(strlen(l_temp)-5))
	    noutimages=noutimages+1
	    if (noutimages > maxfiles) {
	        glogprint( l_logfile, "mprepare", "status", type="error",
		    errno=121, str="Maximum number of output images exceeded \
		    ("//str(maxfiles)//")",verbose+ )
		status=1
		goto clean
	    }
	    out[noutimages]=l_temp//".fits" 
	    if (imaccess(out[noutimages])) {
	        glogprint( l_logfile, "mprepare", "status", type="error",
		    errno=102, str="Output image "//out[noutimages]//" exists",
		    verbose+)
		nbad+=1
	    }
	} # end while
    }

    scanfile=""
    delete(tmpfile,ver-, >& "dev$null")

    # Exit if problems found
    if (nbad > 0) {
        glogprint( l_logfile, "mprepare", "status", type="error", errno=102,
	    str=nbad//" image(s) already exist.",verbose+ )
	status=1
	goto clean
    }

    # if there are too many or too few output images, and any defined
    # at all at this stage - exit with error
    if (nimages!=noutimages && l_outimages!="") {
        glogprint( l_logfile, "mprepare", "status", type="error", errno=121,
	    str="Number of input and output images unequal.",verbose+ )
	status=1
	goto clean
    }

    # If prefix is to be used instead
    if (l_outimages=="" || l_outimages==" ") {
	print(l_prefix) | scan(l_prefix)
	if (l_prefix=="" || l_prefix==" ") {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=121,
	        str="Neither output image name nor output prefix is defined.",
		verbose+ )
	    status=1
	    goto clean
	}
	i=1
	nbad=0
	while (i<=nimages) {
	    out[i]=l_prefix//in[i]//".fits"
	    if (imaccess(out[i])) {
	        glogprint( l_logfile, "mprepare", "status", type="error",
		    errno=102, str="Output image "//out[i]//" already exists.",
		    verbose+)
		nbad+=1
	    }
	    i=i+1
	}
	if (nbad > 0) {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=102,
	        str=nbad//" image(s) already exist.",verbose+)
	    status=1
	    goto clean
	}
    }

    i=1
    while (i<=nimages) { 
	if (l_rawpath=="" || l_rawpath==" ")
	    in[i]=in[i]+".fits"
	else
	    in[i]=l_rawpath//in[i]+".fits"

	# Determine the number of extensions/nod sets.
	header=in[i]//"[0]"
	imgets(header,"MODE", >& "dev$null")
	if (imgets.value == "0") {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=131,
	        str="Could not find the MODE from the primary header.",
		verbose+)
            status=status+1
	    goto nextimage
	}

	modeflag=0
	if (imgets.value == "chop-nod") modeflag=1
	if (imgets.value == "ndchop") modeflag=1
	if (imgets.value == "chop") modeflag=2
	if (imgets.value == "nod") modeflag=3
	if (imgets.value == "ndstare") modeflag=4
	if (imgets.value == "stare") modeflag=4
	if (modeflag == 0) {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=132,
	        str="Unrecognised MODE ("//imgets.value//") in the primary \
		header.",verbose+)
            status=status+1
            goto nextimage
	}

	imgets(header,"NUMEXT", >& "dev$null")
	next_inp=int(imgets.value)


	# Update EXTNAME and EXTVER, assuming all extensions but the MDF 
	# are SCI
	n_j=1
	next_sci = 0
	if ((modeflag == 4) && (next_inp != 1)) {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=123,
	        str="Number of extensions in the primary header not consistent \
		with the MODE (stare).",verbose+)
            status=status+1
            goto nextimage
	}
	if ((modeflag == 2) && (next_inp != 1)) {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=123,
	        str="Number of extensions in the primary header not consistent \
		with the MODE (chop).",verbose+)
            status=status+1
            goto nextimage
	}

	while (n_j<=next_inp) {
	    next_sci = next_sci + 1     
            n_j = n_j+1
	}
	flpr # necessary to work around an IRAF bug 

	if (next_sci > next_inp) {
	    glogprint( l_logfile, "mprepare", "status", type="error", errno=121,
	        str="Maximum number of science extensions exceeded.",
		verbose+ )
	    imdelete (out[i],ver-, >& "dev$null")
	    goto crash
	}

	#----------------------------------------------------------------------
	# start output
	glogprint( l_logfile, "mprepare", "visual", type="visual", 
	    vistype="shortdash", verbose=l_verbose )
	glogprint( l_logfile, "mprepare", "task", type="string",
	    str="    input file ->    output file", verbose=l_verbose )

	#Start on the main mprepare steps - rearrange the order of images 
	#within the extensions, and convert the 4D format to 3D format.

	n_j=1
	next_sci = 0
	
	#Copy PHU into new file
	fxcopy(in[i]//"[0]",out[i], >& "IRAF.log")
	glogprint( l_logfile, "mprepare", "task", type="file", str="IRAF.log",
	    verbose=l_verbose )
	delete("IRAF.log", ver-, >& "dev$null")

	while (n_j<=next_inp) {
	    # Create tmp FITS file name used within this loop
	    tmpim1 = mktemp("tmpim1")
	    tmpim2 = mktemp("tmpim2")
	    tmpim3 = mktemp("tmpim3")
	    tmpim4 = mktemp("tmpim4")
	    
	    if (modeflag == 4) {
		imcopy(in[i]//"["//n_j//"]"//"[*,*,1,1].fits",tmpim1,ver-)      
		fxinsert(tmpim1//".fits",out[i]//"["//next_sci//"]",
		    groups="",ver-)
		glogprint( l_logfile, "mprepare", "engineering", type="string",
		    str="Prepared image for stare "//i,verbose=l_verbose )
		n_j=n_j+1
		
	    } else if (modeflag == 3) {
		imcopy(in[i]//"["//n_j//"]"//"[*,*,1,1].fits",tmpim1,ver-)      
		fxinsert(tmpim1//".fits",out[i]//"["//next_sci//"]",
		    groups="",ver-)
		glogprint( l_logfile, "mprepare", "engineering", type="string",
		    str="Prepared image for nod "//n_j,verbose=l_verbose)
		n_j=n_j+1
		imarith(in[i]//"["//n_j//"]"//"[*,*,1,1].fits","*","-1",tmpim2,ver-)
		fxinsert(tmpim2//".fits",out[i]//"["//next_sci//"]",
		    groups="",ver-)
		glogprint (l_logfile, "mprepare", "engineering", type="string",
		    str="Prepared image for nod "//n_j, verbose=l_verbose)
		n_j=n_j+1
                if  (n_j<next_inp) {
		imarith(in[i]//"["//n_j//"]"//"[*,*,1,1].fits","*","-1",tmpim3,ver-)
		fxinsert(tmpim3//".fits",out[i]//"["//next_sci//"]",
		    groups="",ver-)
		glogprint (l_logfile, "mprepare", "engineering", type="string",
		    str="Prepared image for nod "//n_j, verbose=l_verbose)
		n_j=n_j+1
		imcopy(in[i]//"["//n_j//"]"//"[*,*,1,1].fits",tmpim4,ver-)      
		fxinsert(tmpim4//".fits",out[i]//"["//next_sci//"]",
		    groups="",ver-)
		glogprint (l_logfile, "mprepare", "engineering", type="string",
		    str="Prepared image for nod "//n_j, verbose=l_verbose)
		n_j=n_j+1
		}

	    } else if ((modeflag == 2 || modeflag ==1) && next_inp == 1) {
		#Re-arrange A nod frame
		imcopy(in[i]//"["//n_j//"]"//"[*,*,*,1].fits",tmpim1,ver-)      
		imcopy(in[i]//"["//n_j//"]"//"[*,*,2,1].fits",tmpim1//"[*,*,1]",ver-)      
		imcopy(in[i]//"["//n_j//"]"//"[*,*,3,1].fits",tmpim1//"[*,*,2]",ver-)      
		imcopy(in[i]//"["//n_j//"]"//"[*,*,1,1].fits",tmpim1//"[*,*,3]",ver-,)      
		fxinsert(tmpim1//".fits",out[i]//"["//next_sci//"]",
		    groups="",ver-)
		n_j=n_j+1
		glogprint( l_logfile, "mprepare", "engineering", type="string",
		    str="Rearranged image for chop frame "//i,verbose=l_verbose)
		
	    } else if (modeflag == 1 && next_inp != 1) {
		#Re-arrange A nod frame
		imcopy(in[i]//"["//n_j//"]"//"[*,*,*,1].fits",tmpim1,ver-)      
		imcopy(in[i]//"["//n_j//"]"//"[*,*,2,1].fits",tmpim1//"[*,*,1]",ver-)      
		imcopy(in[i]//"["//n_j//"]"//"[*,*,3,1].fits",tmpim1//"[*,*,2]",ver-)      
		imcopy(in[i]//"["//n_j//"]"//"[*,*,1,1].fits",tmpim1//"[*,*,3]",ver-,)      
		fxinsert(tmpim1//".fits",out[i]//"["//next_sci//"]",
		    groups="",ver-)
		glogprint( l_logfile, "mprepare", "engineering", type="string",
		    str="Rearranged image for nod "//n_j,verbose=l_verbose)
		n_j=n_j+1
		next_sci=next_sci+1
		imdelete(tmpim1,ver-)
		
		tmpim1=mktemp("tmpim1")
		#Re-arrange and invert beam for B nod frame
		imcopy(in[i]//"["//n_j//"]"//"[*,*,*,1].fits",tmpim1,ver-)      
		imcopy(in[i]//"["//n_j//"]"//"[*,*,2,1].fits",tmpim1//"[*,*,2]",ver-)      
		imcopy(in[i]//"["//n_j//"]"//"[*,*,3,1].fits",tmpim1//"[*,*,1]",ver-)      
		imarith(in[i]//"["//n_j//"]"//"[*,*,1,1].fits","*","-1",tmpim2,ver-)
		imcopy(tmpim2,tmpim1//"[*,*,3]",ver-)      
		fxinsert(tmpim1//".fits",out[i]//"["//next_sci//"]",
		    groups="",ver-)
		glogprint( l_logfile, "mprepare", "engineering", type="string",
		    str="Rearranged image for nod "//n_j, verbose=l_verbose )
		n_j=n_j+1
		next_sci=next_sci+1
		
		#Take into account the fact that there might be a 
		#non-divisible by 4 number of nod pairs
		if (n_j < next_inp) {
		    imdelete(tmpim1,ver-)
		    imdelete(tmpim2//".fits",ver-)
		    tmpim1=mktemp("tmpim1")
		    tmpim2=mktemp("tmpim2")
		    
		    #Re-arrange and invert beam for B nod frame
		    imcopy(in[i]//"["//n_j//"]"//"[*,*,*,1].fits",tmpim1,ver-)      
		    imcopy(in[i]//"["//n_j//"]"//"[*,*,2,1].fits",
		        tmpim1//"[*,*,2]",ver-)      
		    imcopy(in[i]//"["//n_j//"]"//"[*,*,3,1].fits",
		        tmpim1//"[*,*,1]",ver-)      
		    imarith(in[i]//"["//n_j//"]"//"[*,*,1,1].fits","*","-1",
		        tmpim2,ver-)
		    imcopy(tmpim2,tmpim1//"[*,*,3]",ver-)      
		    fxinsert(tmpim1//".fits",out[i]//"["//next_sci//"]",
		        groups="",ver-)
		    glogprint( l_logfile, "mprepare", "engineering", 
		        type="string", str="Rearranged image for nod "//n_j,
		        verbose=l_verbose)

		    #Re-arrange A nod frame
		    n_j=n_j+1
		    next_sci=next_sci+1
		    imdelete(tmpim1//".fits",ver-)
		    tmpim1=mktemp("tmpim1")
		    imcopy(in[i]//"["//n_j//"]"//"[*,*,*,1].fits",tmpim1,ver-)      
		    imcopy(in[i]//"["//n_j//"]"//"[*,*,2,1].fits",
		        tmpim1//"[*,*,1]",ver-)      
		    imcopy(in[i]//"["//n_j//"]"//"[*,*,3,1].fits",
		        tmpim1//"[*,*,2]",ver-)      
		    imcopy(in[i]//"["//n_j//"]"//"[*,*,1,1].fits",
		        tmpim1//"[*,*,3]",ver-,)      
		    fxinsert(tmpim1//".fits",out[i]//"["//next_sci//"]",
		        groups="",ver-)
		    glogprint( l_logfile, "mprepare", "engineering",
		        type="string", str="Rearranged image for nod "//n_j,
		        verbose=l_verbose)
		    n_j=n_j+1
		    next_sci=next_sci+1
		}
	    } #end if-else-if on modeflag
	    
	    # Delete tmp images
           imdelete(tmpim1,ver-)
	    if ((modeflag == 2) || (modeflag == 1)) {
	    imdelete(tmpim2,ver-)
           }
	    if (modeflag == 3) {
	    imdelete(tmpim2,ver-)
	    imdelete(tmpim3,ver-)
	    imdelete(tmpim4,ver-)
           }


	} #end while (n_j<=next_inp)

	#-------------
	# Pack up the results and clean up	

	# Time stamps
	date | scan(l_struct)
	gemhedit(out[i]//"[0]","GEM-TLM",l_struct,"Last modification with GEMINI IRAF")
	gemhedit(out[i]//"[0]","MPREPARE",l_struct,"Time stamp for MPREPARE")
	gemhedit(out[i]//"[0]","PREPARE",l_struct,"data have been PREPAREd")
	gemhedit(out[i]//"[0]","DISPAXIS",1,"Dispersion direction")

        glogprint( l_logfile, "mprepare", "visual", type="visual",
	    vistype="empty", verbose=l_verbose )
	glogprint( l_logfile, "mprepare", "engineering", type="string",
	    str="Finished with image number "//i//" at ["//l_struct//"]",
	    verbose=l_verbose )

	n_j=1
	while (n_j <= next_inp) {
	    hedit(out[i]//"["//n_j//"]","AXISLAB4","",add-,addonly-,delete+,
	        verify-,show-,update+)
	    gemhedit(out[i]//"["//n_j//"]","EXTVER",1,"Extension version")
	    gemhedit(out[i]//"["//n_j//"]","EXTNAME","SCI","Extension name")
	    n_j=n_j+1
	}

nextimage:
	i=i+1
    } # end loop

clean:
    if (status==0)
        glogclose( l_logfile, "mprepare", fl_success+, verbose=l_verbose )
    else
        glogclose( l_logfile, "mprepare", fl_success-, verbose=l_verbose )
	
    scanfile="" 
    delete(tmpfile//","//tmphead,ver-, >& "dev$null")

exitnow:
    ;

end
